#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List
from mitosheet.enterprise.api.code_snippets_utils import create_success_return_obj, get_custom_code_snippets
from mitosheet.types import CodeSnippet, StepsManagerType

DEFAULT_CODE_SNIPPETS: List[CodeSnippet] = [
        {
                'Id': 'mito-code-snippet:calculate-correlation-matrix',
                'Name': 'Calculate correlation matrix',
                'Description': 'Compute pairwise correlation of columns, excluding NA/null values.',
                'Code': [
                        "REPLACE_WITH_DATAFRAME_NAME.corr()" 
                ]
        },
        {
                'Id': 'mito-code-snippet:send-email-using-outlook',
                "Name": "Send email using Outlook (Windows)", 
                "Description": "Automatically send email with embedded dataframe",
                "Code": [
                                "import win32com.client",
                                "",
                                "outlook = win32com.client.Dispatch('outlook.application')",
                                "mail = outlook.CreateItem(0)",
                                "mail.To = ['REPLACE_WITH_EMAIL_1', 'REPLACE_WITH_EMAIL_2']",
                                "mail.HTMLBody = REPLACE_WITH_DATAFRAME_NAME.to_html()",
                                "",
                                "# Send the email",
                                "mail.Send()"
                ],
        },
        {
                'Id': 'mito-code-snippet:create-dataframe-from-clipboard',
                'Name': 'Create dataframe from clipboard',
                'Description': 'Read the data that is copied to you clipboard into a pandas dataframe',
                'Code': [
                        "import pandas as pd",
		        "df = pd.read_clipboard()"
                ]
        },
        {
                'Id': 'mito-code-snippet:read-csv-files-from-folder',
                "Name": "Read all CSV files from folder", 
                "Description": "Read the CSV files from a folder, and concate them into a single dataframe.",
                "Code": [
                        "import pandas as pd",
                        "import glob",
                        "import os",
                        "",
                        "path = 'REPLACE_WITH_PATH_TO_FOLDER'",
                        "all_files = glob.glob(os.path.join(path, '*.csv'))",
                        "df = pd.concat((pd.read_csv(f) for f in all_files), ignore_index=True)"
                ]
        },
        {
                'Id': 'mito-code-snippet:diff-between-dataframe',
                "Name": "Find differences between dataframes", 
                "Description": "Subtract one dataframe from another by row",
                "Code": [
                        "df3 = REPLACE_WITH_DATAFRAME_1 - REPLACE_WITH_DATAFRAME_2",
                        "mitosheet.sheet(df3)"
                ]
        } 
]

def get_code_snippets(params: Dict[str, Any], steps_manager: StepsManagerType) -> Dict[str, Any]:
        code_snippets_envs = steps_manager.mito_config.get_code_snippets()

        if code_snippets_envs is None: 
                return create_success_return_obj(DEFAULT_CODE_SNIPPETS)

        mito_config_code_snippets_url = code_snippets_envs["MITO_CONFIG_CODE_SNIPPETS_URL"]

        return get_custom_code_snippets(mito_config_code_snippets_url)
        




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy
import json
import os
from typing import Any, Dict, List, Optional

import pandas as pd
from mitosheet.step_performers.import_steps.excel_range_import import ExcelRangeImportStepPerformer
from mitosheet.step_performers.import_steps.snowflake_import import SnowflakeImportStepPerformer
from mitosheet.types import StepsManagerType
from mitosheet.step_performers.import_steps.simple_import import SimpleImportStepPerformer
from mitosheet.step_performers.import_steps.excel_import import ExcelImportStepPerformer
from mitosheet.step_performers.import_steps.dataframe_import import DataframeImportStepPerformer
from mitosheet.step_performers.import_steps import is_import_step_type

def get_sublist_at_index_from_optional_list(l: Optional[List[Any]], index: int) -> Optional[List[Any]]:
    if l is None:
        return None
        
    return [l[index]]


def get_import_data_with_single_import_list(step_type: str, params: Dict[str, Any]) -> List[Dict[str, Any]]:
    """
    Given a step_type that is an import, and the parameters to that import step,
    this function will create a list of DataframeCreationData that represent
    parameters to steps that only create a single dataframe.

    This is useful because the frontend allows users to update dataframe imports one 
    be one - and so having each step only do a single import is convenient for working
    with the data on the frontend.
    """
    if step_type == SimpleImportStepPerformer.step_type():
        return [{
            'step_type': step_type,
            'params': {
                'file_names': [file_name],
                'delimeters': get_sublist_at_index_from_optional_list(params.get('delimeters', None), index),
                'encodings': get_sublist_at_index_from_optional_list(params.get('encodings', None), index),
                'decimals': get_sublist_at_index_from_optional_list(params.get('decimals', None), index),
                'skiprows': get_sublist_at_index_from_optional_list(params.get('skiprows', None), index),
                'error_bad_lines': get_sublist_at_index_from_optional_list(params.get('error_bad_lines', None), index)
            }
        } for index, file_name in enumerate(params['file_names'])]

    if step_type == ExcelImportStepPerformer.step_type():
        return [{
            'step_type': step_type,
            'params': {
                'file_name': params['file_name'],
                'sheet_names': [sheet_name],
                'has_headers': params['has_headers'],
                'skiprows': params['skiprows'],
                'decimal': params['decimal']
            }
        } for sheet_name in params['sheet_names']]

    if step_type == DataframeImportStepPerformer.step_type():
        return [{
            'step_type': step_type,
            'params': {
                'df_names': [df_name],
            }
        } for df_name in params['df_names']]

    if step_type == ExcelRangeImportStepPerformer.step_type():
        return [{
            'step_type': step_type,
            'params': {
                'file_path': params['file_path'],
                'sheet_name': params['sheet_name'],
                'range_imports': [range_import]
            }
        } for range_import in params['range_imports']]

    if step_type == SnowflakeImportStepPerformer.step_type():
        return [{
            'step_type': step_type,
            'params': {
                'table_loc_and_warehouse': params['table_loc_and_warehouse'],
                'query_params': params['query_params']
            }
        }]

    return []


def get_imported_files_and_dataframes_from_current_steps(params: Dict[str, Any], steps_manager: StepsManagerType) -> List[Dict[str, Any]]:
    """
    Returns a list all the imported files and dataframes, and their import params. Does so by turning
    the import into a list of imports that only import a single dataframe, as this is convenient
    for what we work with on the frontend.

    [
        {
            step_id: string,
            imports: ({
                step_type: 'simple_import'
                params: CSVImportParams
            } |
            {
                step_type: 'excel_import'
                params: ExcelImportParams
            } |
            {
                step_type: 'dataframe_import'
                params: DataframeImportParams
            } |
            {
                step_type: 'excel_range_import'
                params: ExcelRangeImportParams
            } |
            {
                step_type: 'snowflake_import'
                params: SnowflakeImportParams
            }  
            )[]
        }
	]

    NOTE: see comment in get_imported_files_and_dataframes_from_analysis_name for additional context. 
    This function is only used once an analysis exists, in which case the step ids are used for 
    matching imports. Note that results from this function will lead to changing imports with the `update_imports`
    update event -- while the get_imported_files_and_dataframes_from_analysis_name function will change imports
    through a parameter to the replay_analysis update.
    """

    # Then, we turn import steps into a version of the step that just creates a single dataframe
    # as this is what is easiest to work with on the frontend
    import_steps_with_just_one_dataframe = []
    for step in steps_manager.steps_including_skipped:
        if is_import_step_type(step.step_type):
            import_steps_with_just_one_dataframe.append({
                'step_id': step.step_id,
                'imports': get_import_data_with_single_import_list(step.step_type, step.params)
            })

    # Then, turn this into the output format
    return import_steps_with_just_one_dataframe




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.filter_code_chunk import FAKE_COLUMN_HEADER, get_entire_filter_string
from mitosheet.is_type_utils import is_int_dtype, is_float_dtype
from mitosheet.state import NUMBER_FORMAT_ACCOUNTING, NUMBER_FORMAT_CURRENCY, NUMBER_FORMAT_PERCENTAGE, NUMBER_FORMAT_PLAIN_TEXT, NUMBER_FORMAT_SCIENTIFIC_NOTATION, State
from mitosheet.transpiler.transpile_utils import TAB, column_header_list_to_transpiled_code, column_header_to_transpiled_code
from mitosheet.types import ColumnFormat, ColumnHeader
from mitosheet.utils import MAX_ROWS
from mitosheet.pro.conditional_formatting_utils import get_conditonal_formatting_result

OPEN_BRACKET = "{"
CLOSE_BRACKET = "}"

def get_format_string_for_column_format(column_format: Optional[ColumnFormat], dtype: str) -> Optional[str]:
    """For a given column format, returns a format string or None if it's the default"""
    if column_format is None:
        return None

    type = column_format.get('type', None)
    precision = column_format.get('precision', None)

    if precision is not None:
        precision_string = f'.{precision}'
    else:
        # If there is no precision, we use the default precision for float to 2
        if is_float_dtype(dtype):
            precision_string = '.2'
        else:
            precision_string = ''

    # If this is an int column, and we have no decimals, then we use the d format as to not get any decimals
    format_f_or_d = 'f'
    if is_int_dtype(dtype) and precision_string == '':
        format_f_or_d = 'd'

    if type == NUMBER_FORMAT_PLAIN_TEXT:
        return f"\"{OPEN_BRACKET}:{precision_string}{format_f_or_d}{CLOSE_BRACKET}\""
    elif type == NUMBER_FORMAT_CURRENCY:
        return f"\"${OPEN_BRACKET}:{precision_string}{format_f_or_d}{CLOSE_BRACKET}\""    
    elif type == NUMBER_FORMAT_ACCOUNTING:
        return f"lambda val: '${OPEN_BRACKET}:>,{precision_string}{format_f_or_d}{CLOSE_BRACKET}'.format(abs(val)) if val > 0 else '$({OPEN_BRACKET}:>,{precision_string}{format_f_or_d}{CLOSE_BRACKET})'.format(abs((val)))"
    elif type == NUMBER_FORMAT_PERCENTAGE:
        # For some reason, if the int dype is used, then we need to add a leading .0 on the percentage
        # so we don't display decimals
        if is_int_dtype(dtype) and precision_string == '':
            precision_string = '.0'

        return f"\"{OPEN_BRACKET}:,{precision_string}%{CLOSE_BRACKET}\""
    elif type == NUMBER_FORMAT_SCIENTIFIC_NOTATION:
        return f"\"{OPEN_BRACKET}:{precision_string}E{CLOSE_BRACKET}\""

    # If we have no formatting, we apply the default formatting with the precision (if the precision is set)
    if precision is not None:
        return f"\"{OPEN_BRACKET}:{precision_string}{format_f_or_d}{CLOSE_BRACKET}\""

    # Otherwise, we return None
    return None

def get_all_columns_format_code(state: State, sheet_index: int) -> Optional[str]:
    """Returns the .format call for all of the columns"""
    df_format = state.df_formats[sheet_index]
    df = state.dfs[sheet_index]
    columns = df_format['columns']

    # Store all the ColumnHeaders under their format string, so that we can easily
    # combine all of them together, and then use the .format efficiently
    format_string_to_column_headers: Dict[str, List[ColumnHeader]] = {}
    for column_id, column_format in columns.items():
        column_header = state.column_ids.get_column_header_by_id(sheet_index, column_id)
        dtype = str(df[column_header].dtype)
        format_string = get_format_string_for_column_format(column_format, dtype)

        if format_string is not None:
            if format_string not in format_string_to_column_headers:
                format_string_to_column_headers[format_string] = []
            format_string_to_column_headers[format_string].append(column_header)

    # Combine this all into one format string
    all_columns_format_code = ''
    for format_string, column_headers in format_string_to_column_headers.items():
        all_columns_format_code += f'.format({format_string}, subset={column_header_list_to_transpiled_code(column_headers)})'

    if len(all_columns_format_code) > 0:
        return all_columns_format_code

    return None

def get_transpiled_table_style(selector: str, props: List[Tuple[str, Optional[str]]]) -> Optional[str]:
    """A helper that returns a selector and props pair to be used in a set_table_styles call"""
    # We filter out all the props that have None as values
    props = [prop for prop in props if prop[1] is not None]
    if len(props) > 0:
        return f"{OPEN_BRACKET}'selector': \'{selector}\', 'props': {column_header_list_to_transpiled_code(props)}{CLOSE_BRACKET}"
    return None

def get_headers_format_code(state: State, sheet_index: int) -> Optional[str]:
    """Returns header formatting code"""
    df_format = state.df_formats[sheet_index]
    headers = df_format['headers']
    
    return get_transpiled_table_style(
        'thead', 
        [
            ('color', headers.get('color', None)),
            ('background-color', headers.get('backgroundColor', None))
        ]
    )

def get_rows_format_code(state: State, sheet_index: int, even_or_odd: str) -> Optional[str]:
    """Returns row formatting code"""
    df_format = state.df_formats[sheet_index]
    rows = df_format['rows']

    # By default, Pandas indexes rows starting with index 0. 
    # However, when we apply the pandas styler, it defaults to calculating if rows
    # are even or odd starting at index 1. Therefore, to keep them in sync, we reverse
    # the even or odd substring of the css selector. 
    even_or_odd_css_selector = 'even' if even_or_odd == 'odd' else 'odd'

    evenOrOdd = rows.get(even_or_odd, dict())
    return get_transpiled_table_style(
        f'tbody tr:nth-child({even_or_odd_css_selector})',
        [
            ('color', evenOrOdd.get('color', None)),
            ('background-color', evenOrOdd.get('backgroundColor', None))
        ]
    )

def get_border_format_code(state: State, sheet_index: int) -> Optional[str]:
    """Returns border formatting code"""
    df_format = state.df_formats[sheet_index]
    border = df_format['border']

    borderStyle = border.get('borderStyle', 'none')
    borderColor = border.get('borderColor', 'black')

    # If there is no border, then don't return anything
    if borderStyle == 'none':
        return None
    
    border_string = f'1px {borderStyle} {borderColor}'

    return get_transpiled_table_style(
        '',
        [
            ('border', border_string),
        ]
    )

def get_table_styles_code(state: State, sheet_index: int) -> Optional[str]:
    """Returns the call to set_table_styles"""
    table_styles_code = ''
    header_format_code = get_headers_format_code(state, sheet_index)
    even_format_code = get_rows_format_code(state, sheet_index, 'even')
    odd_format_code = get_rows_format_code(state, sheet_index, 'odd')
    border_format_code = get_border_format_code(state, sheet_index)

    if header_format_code is not None:
        table_styles_code += f'{TAB}{TAB}{header_format_code},\n'
    if even_format_code is not None:
        table_styles_code += f'{TAB}{TAB}{even_format_code},\n'
    if odd_format_code is not None:
        table_styles_code += f'{TAB}{TAB}{odd_format_code},\n'
    if border_format_code is not None:
        table_styles_code += f'{TAB}{TAB}{border_format_code},\n'

    if len(table_styles_code) > 0:
        return f".set_table_styles([\n{table_styles_code}])"
    return None


def get_conditional_format_code_list(state: State, sheet_index: int) -> Tuple[Optional[List[str]], Optional[bool]]:
    """Returns all the code to set the conditional formats"""
    df_name = state.df_names[sheet_index]
    df = state.dfs[sheet_index]
    conditional_formats = state.df_formats[sheet_index]['conditional_formats']

    # We get the conditional formatting results, and we filter out any columns that are 
    # are invalid with the filters that are applied
    conditional_formatting_result = get_conditonal_formatting_result(
        state,
        sheet_index,
        df,
        conditional_formats
    )

    all_code = []
    uses_numpy = False
    for conditional_format in conditional_formats:
        formatUUID = conditional_format['format_uuid']
        filters = conditional_format['filters']
        column_ids = conditional_format['columnIDs']
        color = conditional_format.get('color', None)
        background_color = conditional_format.get('backgroundColor', None)

        final_column_ids = list(set(column_ids).difference(conditional_formatting_result['invalid_conditional_formats'].get(formatUUID, [])))

        if len(final_column_ids) == 0:
            continue

        # Get the column headers
        column_headers = state.column_ids.get_column_headers_by_ids(sheet_index, final_column_ids)
        transpiled_column_headers = column_header_list_to_transpiled_code(column_headers)

        entire_filter_string = get_entire_filter_string(state, sheet_index, 'And', filters)
        if entire_filter_string is None:
            continue

        # The filter string, if it's not given a column header when generated, uses this FAKE_COLUMN_HEADER. 
        # See get_entire_filter_string for further description, but it allows us to filter the series var used
        # in the conditional formatting implementation
        entire_filter_string = entire_filter_string.replace(f'{df_name}[{column_header_to_transpiled_code(FAKE_COLUMN_HEADER)}]', "series")

        color_string = ''
        if color is not None:
            color_string += f'color: {color}'
        if background_color is not None:
            if len(color_string) > 0:
                color_string += '; '
            color_string += f'background-color: {background_color}'

        if len(color_string) > 0:
            uses_numpy = True
            all_code.append(f".apply(lambda series: np.where({entire_filter_string}, '{color_string}', None), subset={transpiled_column_headers})")

    if len(all_code) > 0:
        return all_code, uses_numpy

    return None, None

def check_conditional_filters_have_filter_condition_that_requires_whole_dataframe(state: State, sheet_index: int) -> bool:
    """
    Returns true if any of the conditional formats have any filter conditions that require
    the full dataframe to be present to calculate correctly.
    """
    from mitosheet.step_performers.filter import check_filters_contain_condition_that_needs_full_df
    conditional_formats = state.df_formats[sheet_index]['conditional_formats']

    for conditional_format in conditional_formats:
        filters = conditional_format['filters']
        if check_filters_contain_condition_that_needs_full_df(filters):
            return True

    return False


def get_dataframe_format_code(state: State, sheet_index: int) -> Tuple[Optional[str], Optional[bool]]:
    """Returns all the code to set the df_formatting on the dataframe from the state."""
    df_name = state.df_names[sheet_index]
    df = state.dfs[sheet_index]

    if len(df) <= MAX_ROWS: 
        dataframe_format_string = f"{df_name}_styler = {df_name}.style"
    else: 
        # If there are more than the max rows, we don't display all of them
        dataframe_format_string = f"{df_name}_styler = {df_name}.head({MAX_ROWS}).style"

        # If there is a .head call, and we have filter conditions that require access to the entire
        # dataframe, than we generate an extra comment to let the user know something might be incorrect
        if check_conditional_filters_have_filter_condition_that_requires_whole_dataframe(state, sheet_index):
            dataframe_format_string = f'# This .head call avoids printing too much data, but may lead to incorrectly calculated conditional formats\n{dataframe_format_string}'
        

    format_code = [
        get_all_columns_format_code(state, sheet_index),
        get_table_styles_code(state, sheet_index),
    ]
    conditional_format_code, uses_numpy = get_conditional_format_code_list(state, sheet_index)
    if conditional_format_code:
        format_code += conditional_format_code

    # If all the format code is None, then we write nothing
    if all(map(lambda x: x is None, format_code)):
        return None, None
    
    for line in format_code:
        if line is None:
            continue
        dataframe_format_string += f"\\\n{TAB}{line}"
    
    return dataframe_format_string, uses_numpy

class SetDataframeFormatCodeChunk(CodeChunk):
    """
    This is the first postprocessing code chunk. Postprocessing code chunks are executed after all other code chunks,
    and are given access to the final state. They have no parameters otherwise.
    """

    def get_display_name(self) -> str:
        return 'Set Dataframe Format'
    
    def get_description_comment(self) -> str:
        return f'Formatted dataframes. View these styling objects to see the formatted dataframe'

    def get_code(self) -> Tuple[List[str], List[str]]:
        code = []
        uses_numpy = False
        for sheet_index in range(len(self.post_state.df_formats)):
            dataframe_format_code, _uses_numpy = get_dataframe_format_code(self.post_state, sheet_index)
            if dataframe_format_code is not None:
                code.append(dataframe_format_code)
            if _uses_numpy:
                uses_numpy = True

        imports = ['import numpy as np'] if uses_numpy else []

        # Make sure to import numpy if we use it
        return code, imports



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import List

from mitosheet.user import set_user_field
from mitosheet.user.schemas import UJ_MITOSHEET_PRO, UJ_MITOSHEET_TELEMETRY

GO_PRO_EVENT = 'go_pro'
GO_PRO_PARAMS: List[str] = []

def execute_go_pro_update(steps_manager):
    set_user_field(UJ_MITOSHEET_PRO, True)
    set_user_field(UJ_MITOSHEET_TELEMETRY, False)

GO_PRO_UPDATE = {
    'event_type': GO_PRO_EVENT,
    'params': GO_PRO_PARAMS,
    'execute': execute_go_pro_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains all functions that are useful for control flow. For now, this
is just IF statements.

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""
import functools

import pandas as pd
from mitosheet.public.v1.sheet_functions.sheet_function_utils import (
    fill_series_with_one_index, try_extend_series_to_index)
from mitosheet.public.v1.sheet_functions.types.decorators import (
    convert_arg_to_series_type, convert_args_to_series_type, fill_nans,
    filter_nans, handle_sheet_function_errors)


@handle_sheet_function_errors
@filter_nans
@convert_args_to_series_type('bool', 'skip', ('default', True))
def AND(*argv: pd.Series) -> pd.Series:
    """
    {
        "function": "AND",
        "description": "Returns True if all of the provided arguments are True, and False if any of the provided arguments are False.",
        "search_terms": ["and", "&", "if", "conditional"],
        "examples": [
            "AND(True, False)",
            "AND(Nums > 100, Nums < 200)",
            "AND(Pay > 10, Pay < 20, Status == 'active')"
        ],
        "syntax": "AND(boolean_condition1, [boolean_condition2, ...])",
        "syntax_elements": [{
                "element": "boolean_condition1",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            },
            {
                "element": "boolean_condition2 ... [OPTIONAL]",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            }
        ]
    }
    """
    argv = fill_series_with_one_index(argv)

    return functools.reduce((lambda x, y: x & y), argv)



@handle_sheet_function_errors
@fill_nans(0, False)
@convert_arg_to_series_type(
    0,
    'bool', 
    'error', 
    ('default', True)
)
def BOOL(series: pd.Series) -> pd.Series:
    """
    {
        "function": "BOOL",
        "description": "Converts the passed arguments to boolean values, either True or False. For numberic values, 0 converts to False while all other values convert to True.",
        "search_terms": ["bool", "boolean", "true", "false", "dtype", "convert"],
        "examples": [
            "BOOL(Amount_Payed)",
            "AND(BOOL(Amount_Payed), Is_Paying)"
        ],
        "syntax": "BOOL(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "An series to convert to boolean values, either True or False."
            }
        ]
    }
    """
    return series


@handle_sheet_function_errors
@convert_arg_to_series_type(
    0,
    'bool',
    'error',
    'error'
)
@convert_arg_to_series_type(
    1,
    'series',
    'error',
    'error',
)
@convert_arg_to_series_type(
    2,
    'series',
    'error',
    'error',
)
def IF(condition: pd.Series, true_series: pd.Series, false_series: pd.Series) -> pd.Series:
    """
    {
        "function": "IF",
        "description": "Returns one value if the condition is True. Returns the other value if the conditon is False.",
        "search_terms": ["if", "conditional", "and", "or"],
        "examples": [
            "IF(Status == 'success', 1, 0)",
            "IF(Nums > 100, 100, Nums)",
            "IF(AND(Grade >= .6, Status == 'active'), 'pass', 'fail')"
        ],
        "syntax": "IF(boolean_condition, value_if_true, value_if_false)",
        "syntax_elements": [{
                "element": "boolean_condition",
                "description": "An expression or series that returns True or False values. Valid conditions for comparison include ==, !=, >, <, >=, <=."
            },
            {
                "element": "value_if_true",
                "description": "The value the function returns if condition is True."
            },
            {
                "element": "value_if_false",
                "description": "The value the function returns if condition is False."
            }
        ]
    }
    """
    true_series = try_extend_series_to_index(true_series, condition.index)
    false_series = try_extend_series_to_index(false_series, condition.index)

    return pd.Series(
        data=[true_series.loc[i] if c else false_series.loc[i] for i, c in condition.items()],
        index=condition.index
    )


@handle_sheet_function_errors
@filter_nans
@convert_args_to_series_type('bool', 'skip', ('default', True))
def OR(*argv: pd.Series) -> pd.Series:
    """
    {
        "function": "OR",
        "description": "Returns True if any of the provided arguments are True, and False if all of the provided arguments are False.",
        "search_terms": ["or", "if", "conditional"],
        "examples": [
            "OR(True, False)",
            "OR(Status == 'success', Status == 'pass', Status == 'passed')"
        ],
        "syntax": "OR(boolean_condition1, [boolean_condition2, ...])",
        "syntax_elements": [{
                "element": "boolean_condition1",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            },
            {
                "element": "boolean_condition2 ... [OPTIONAL]",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            }
        ]
    }
    """

    argv = fill_series_with_one_index(argv)
    return functools.reduce((lambda x, y: x | y), argv)



# TODO: we should see if we can list these automatically!
CONTROL_FUNCTIONS = {
    'AND': AND,
    'BOOL': BOOL,
    'IF': IF,
    'OR': OR,
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from copy import copy
import re
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import pandas as pd
import numpy as np
from mitosheet.types import CodeOptions, ColumnHeader, ParamName, ParamValue, StepsManagerType
from mitosheet.utils import is_prev_version

# TAB is used in place of \t in generated code because
# Jupyter turns \t into a grey arrow, but converts four spaces into a tab.
TAB = '    '
NEWLINE_TAB = f'\n{TAB}'
NEWLINE = '\n'
OPEN_BRACKET = "{"
CLOSE_BRACKET = "}"


def column_header_list_to_transpiled_code(column_headers: Union[List[ColumnHeader], Set[ColumnHeader], List[Tuple[str, Optional[str]]]]) -> str:
    """
    A helper function for turning a list of column headers into a 
    valid list of Python code.
    """
    transpiled_column_headers = [
        column_header_to_transpiled_code(column_header)
        for column_header in column_headers
    ]
    joined_transpiled_column_headers = ', '.join(transpiled_column_headers)
    return f'[{joined_transpiled_column_headers}]'


def column_header_to_transpiled_code(column_header: ColumnHeader, tab_level: int=0) -> str:
    """
    Makes sure the column header is correctly transpiled to 
    code in a way that makes sure it's referenced properly.

    Handles multi-index, boolean, string, and number columns 
    correctly.
    """
    # If this is a multi-index header, then we turn each of the pieces of the column
    # header into valid transpiled code, and then we combine them into a tuple
    if isinstance(column_header, tuple):
        column_header_parts = [column_header_to_transpiled_code(column_header_part) for column_header_part in column_header]
        column_header_parts_joined = ', '.join(column_header_parts)
        return f'({column_header_parts_joined})'
    if isinstance(column_header, list):
        column_header_parts = [column_header_to_transpiled_code(column_header_part, tab_level=tab_level+1) for column_header_part in column_header]
        column_header_parts_joined = f',\n{TAB*(tab_level + 1)}'.join(column_header_parts)
        return f'[\n{TAB*(tab_level + 1)}{column_header_parts_joined}\n{TAB*tab_level}]'

    # We must handle np.nan first because isinstance(np.nan, float) evaluates to True
    if not is_prev_version(pd.__version__, '1.0.0') and column_header is np.nan:
        return 'pd.np.nan'
    elif isinstance(column_header, int) or isinstance(column_header, float) or isinstance(column_header, bool):
        return str(column_header)
    elif isinstance(column_header, pd.Timestamp):
        return f'pd.to_datetime(\'{column_header.strftime("%Y-%m-%d %X")}\')'
    elif isinstance(column_header, pd.Timedelta):
        return f'pd.to_timedelta(\'{str(column_header)}\')'
    elif column_header is pd.NaT:
        return 'pd.NaT'
    elif not is_prev_version(pd.__version__, '1.0.0') and column_header is pd.NA:
        return 'pd.NA'
    elif isinstance(column_header, str) and "'" in column_header:
        return f"\"{column_header}\""

    return repr(column_header)

def list_to_string_without_internal_quotes(list: List[Any]) -> str:
    """
    Helper function for formatting a list as a string without 
    leading and trailing '
    """
    string = (', ').join(list)
    return "[" + string +  "]"

def column_header_map_to_string(column_header_map: Dict[ColumnHeader, ColumnHeader]) -> str:
    if len(column_header_map) <= 3:
        # If there are only a few column headers, we put them in a single line
        result = '{' 
        for column_header_key, column_header_value in column_header_map.items():
            result += f'{column_header_to_transpiled_code(column_header_key)}: {column_header_to_transpiled_code(column_header_value)}, '
        result = result[:-2] + "}" # don't take the last comma and space
        return result
    else:
        result = '{\n' 
        for column_header_key, column_header_value in column_header_map.items():
            result += f'{TAB}{column_header_to_transpiled_code(column_header_key)}: {column_header_to_transpiled_code(column_header_value)},\n'
        result = result[:-2] + "\n}" # don't take the last comma and new line
        return result


def param_dict_to_code(param_dict: Dict[str, Any], level: int=0, as_single_line: bool=False, tab_level: int=1, is_dict_entry: bool=False) -> str:
    """
    Takes a potentially nested params dictonary and turns it into a
    code string that we can use in the graph generated code.

    level should be 0 if we are at the highest level dict, and otherwise
    should increment by 1 anytime we enter a new subdictonary.
    """

    # Make sure we handle as a single line properly
    if as_single_line:
        TAB_CONSTANT = ''
        NEWLINE_CONSTANT = ''
    else:
        TAB_CONSTANT = TAB
        NEWLINE_CONSTANT = f'\n{TAB_CONSTANT * tab_level}'

    code = f"{NEWLINE_CONSTANT}"

    value_num = 0
    for key, value in param_dict.items():
        # If we're defining a dict entry, we need to add quotes around the key and use a colon
        key_definition = f'"{key}": ' if is_dict_entry else f'{key}='
        if isinstance(value, dict):
            # Recurse on this nested param dictonary
            code_chunk = f"{key_definition}{{{f'{param_dict_to_code(value, level=level + 1, is_dict_entry=True)}{NEWLINE_CONSTANT}{TAB_CONSTANT * (level+1)}'}}}"
        elif isinstance(value, list):
            code_chunk = f"{key_definition}{column_header_to_transpiled_code(value, tab_level=tab_level + 1)}"
        else:
            # We use this slighly misnamed function to make sure values get transpiled right
            code_chunk = f"{key_definition}{column_header_to_transpiled_code(value)}"
        
        # If we're not on the first value in this dict, we need to add a 
        # command new line after the last value
        if value_num != 0:
            code += f", {NEWLINE_CONSTANT}"

        value_num += 1

        # Add spacing before the param
        code += f"{TAB_CONSTANT * (level + 1)}"

        code += f"{code_chunk}"

    if level == 0:
        code += f"{NEWLINE_CONSTANT}"
    
    return code

def get_str_param_name(steps_manager: StepsManagerType, index: int) -> str:
    # We go and find the first state, and then get the name of the df at this index
    df_name = steps_manager.steps_including_skipped[0].final_defined_state.df_names[index]
    return df_name + '_path'

def _get_params_dict_for_function_call(steps_manager: StepsManagerType, function_params: Dict[ParamName, ParamValue]) -> Dict[ParamName, ParamValue]:
    """
    Returns a dictionary of the param names and values for the function call. Takes special care
    to when you pass function_params that overwrite some of the original args to the function call.
    """
    from mitosheet.updates.args_update import is_string_arg_to_mitosheet_call

    original_args_values = steps_manager.original_args_raw_strings
    original_args_names = []

    for index, original_arg_value in enumerate(original_args_values):

        if original_arg_value in function_params.values():
            # If this arg is being overwritten by a param, we need to get the name of the param
            # instead of the original arg name
            param_name = list(function_params.keys())[list(function_params.values()).index(original_arg_value)]
            original_args_names.append(param_name)
        else:
            if is_string_arg_to_mitosheet_call(original_arg_value):
                original_args_names.append(get_str_param_name(steps_manager, index))
            else:
                original_args_names.append(original_arg_value)

    args_dict = {param_name: param_value for param_name, param_value in zip(original_args_names, original_args_values)}

    # Then, we extend with the new params, not including any params that are already in the args
    for param_name, param_value in function_params.items():
        if param_name not in args_dict:
            args_dict[param_name] = param_value

    return args_dict


def _get_param_names_string(steps_manager: StepsManagerType, function_params: Dict[ParamName, ParamValue]) -> str:
    params_names = list(_get_params_dict_for_function_call(steps_manager, function_params).keys())
    return ", ".join(params_names)

def _get_param_values_string(steps_manager: StepsManagerType, function_params: Dict[ParamName, ParamValue]) -> str:
    params_values = list(_get_params_dict_for_function_call(steps_manager, function_params).values())
    return ", ".join(params_values)

def _get_return_variables_string(steps_manager: StepsManagerType, function_params: Dict[ParamName, ParamValue]) -> str:
     
    final_df_names = copy(steps_manager.curr_step.df_names)
    for param_name, param_value in function_params.items():
        if param_value in final_df_names:
            final_df_names[final_df_names.index(param_value)] = param_name

    return ", ".join(final_df_names)

import re

def replace_newlines_with_newline_and_tab(text: str) -> str:
    pattern = r'(?<!\\)\n'  # Negative lookbehind for '\'
    replacement = '\n' + f'{TAB}'  # Newline followed by a tab
    result = re.sub(pattern, replacement, text)
    return result

def convert_script_to_function(
        steps_manager: StepsManagerType, 
        imports: List[str], 
        code: List[str], 
        function_name: str, 
        function_params: Dict[ParamName, ParamValue],
        call_function: bool
    ) -> List[str]:
    """
    Given a list of code lines, puts it inside of a function.
    """
    final_code = []

    # Add the imports
    final_code += imports
    final_code.append("")

    # The param
    param_names = _get_param_names_string(steps_manager, function_params)

    # Add the function definition
    final_code.append(f"def {function_name}({param_names}):")

    for line in code:
        # Add the code, making sure to indent everything, even if it's on the newline
        # or if it's the closing paren. We take special care not to mess inside of any strings, simply
        # by indenting any newline that is not preceeded by a \
        line = f"{TAB}{line}"
        line = replace_newlines_with_newline_and_tab(line)

        # Then, for any additional function params we defined, we relace the internal param value. Note that 
        # we only replace for 
        for param_name, param_value in function_params.items():
            if "'" in param_value or '"' in param_value:
                line = line.replace(param_value, param_name)
            else:
                line = re.sub(r'\b%s\b' % re.escape(param_value), param_name, line) 

        final_code.append(line)

    # Add the return statement, where we return the final dfs
    return_variables_string = _get_return_variables_string(steps_manager, function_params)
    final_code.append(f"{TAB}return {return_variables_string}")
    final_code.append("")

    # Build the params and variables taking special care to ensure that dataframes and file paths 
    # that are passed as parameters to the function. 
    final_params_to_call_function_with = []

    for param_name, param_value in _get_params_dict_for_function_call(steps_manager, function_params).items():
        if param_name in function_params:
            final_params_to_call_function_with.append(param_name)
            final_code.append(f"{param_name} = {param_value}")
        else:
            final_params_to_call_function_with.append(param_value)

    if len(function_params) > 0:
        final_code.append("")

    # If we are not calling the function, we just return the code without the call at the end
    if not call_function:
        return final_code

    final_params_to_call_function_with_string = ", ".join(final_params_to_call_function_with)

    if len(return_variables_string) > 0:
        final_code.append(f"{return_variables_string} = {function_name}({final_params_to_call_function_with_string})")
    else:
        final_code.append(f"{function_name}({final_params_to_call_function_with_string})")

    return final_code



def get_default_code_options(analysis_name: str) -> CodeOptions:
    return {
        'as_function': False,
        'call_function': True,
        'function_name': 'function_' + analysis_name[-4:], # Give it a random name, just so we don't overwrite them
        'function_params': dict()
    }


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

"""
The mitosheet package is distributed under the mitosheet and mitosheet3
package names on pip. The package.json will tell you all you need to 
know about which one we are in currently using.

As such, this setup.py script reads in the package.json and sets up
the proper package.
"""

from __future__ import print_function
from glob import glob
from os.path import join as pjoin
import json
import setuptools
from pathlib import Path


from jupyter_packaging import (
    create_cmdclass,
    install_npm,
    ensure_targets,
    combine_commands,
    skip_if_exists,
)

from setuptools import setup

HERE = Path(__file__).parent.resolve()

package_json = json.loads(open('package.json').read())
lab_path = Path(pjoin(HERE, 'mitosheet', 'labextension'))
notebook_path = Path(pjoin(HERE, 'mitosheet', 'nbextension'))

python_requires='>=3.4'

name = 'mitosheet'

# Representative files that should exist after a successful build
jstargets = [
    str(lab_path / "package.json"),
    str(notebook_path / "index.js"),
]

package_data_spec = {
    'mitosheet': ["*"],
}

labext_name = name

data_files_spec = [
    # Notebook extension data files
    ('share/jupyter/nbextensions/mitosheet', notebook_path, '**'),
    ('etc/jupyter/nbconfig/notebook.d', '.', 'mitosheet.json'),

    # Lab extension data files
    ("share/jupyter/labextensions/%s" % labext_name, str(lab_path), "**"),
    ("share/jupyter/labextensions/%s" % labext_name, str(HERE), "install.json"),
]


cmdclass = create_cmdclass("jsdeps",
    package_data_spec=package_data_spec,
    data_files_spec=data_files_spec
)

js_command = combine_commands(
    install_npm(HERE, build_cmd="build:all", npm=["jlpm"]),
    ensure_targets(jstargets),
)

is_repo = (HERE / ".git").exists()
if is_repo:
    cmdclass["jsdeps"] = js_command
else:
    cmdclass["jsdeps"] = skip_if_exists(jstargets, js_command)

setup_args = dict(
    name                    = name,
    version                 = package_json["version"],
    url                     = package_json["homepage"],
    author                  = package_json["author"]["name"],
    author_email            = package_json["author"]["email"],
    description             = package_json["description"],
    license                 = "GNU Affero General Public License v3",
    long_description="""
    To learn more about Mito, checkout out our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
    Before installing Mito \n\n
    1. Check that you have Python 3.6 or above. To check your version of Python, open a new terminal, and type python3 --version. If you need to install or update Python, restart your terminal after doing so.\n\n
    2. Checkout our terms of service and privacy policy. By installing Mito, you're agreeing to both of them. Please contact us at aaron [@] sagacollab [dot] com with any questions.\n\n
    Installation Instructions \n\n
    For more detailed installation instructions, see our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
    1. pip install mitosheet\n\n
    2. Launch JupyterLab 3.0 and open a new notebook\n\n
    3. In the notebook, run the following code:\n\n
    import mitosheet\n\n
    mitosheet.sheet()\n\n
    """,
    long_description_content_type = "text/markdown",
    cmdclass                 = cmdclass,
    packages                 = setuptools.find_packages(exclude=['deployment']),
    package_data             = {'': ['*.js', '*.css', '*.html']},
    install_requires=[        
        "jupyterlab~=3.0",
        # We allow users to have many versions of pandas installed. All functionality should
        # work, with the exception of Excel import, which might require additonal dependencies
        'pandas>=0.24.2',
        'analytics-python',
        # Graphing libraries
        'plotly>=4.14.3',
        'chardet>=3.0.4',
        # For XLSX, reading - we don't fix so works on all python versions
        'openpyxl',
        # xlsxwriter is needed for adding formatting to exported Excel sheets. 
        # We pin to a pretty old version because the formatting functionality hasn't changed in a long time.
        'xlsxwriter>=0.6.9,<=3.0.2',
    ],
    extras_require = {
        'test': [
            'pytest',
            'flake8',
            'types-chardet',
            'types-requests',
            'mypy',
            'pytest_httpserver'
        ],
        'deploy': [
            'wheel', 
            'twine',
            "jupyter_packaging<=0.10.6",
            "setuptools==56.0.0"
        ],
        'streamlit': [
            'streamlit>=1.0',
        ],
        'optional_feature_dependencies': [
            # According to this documentation (https://github.com/snowflakedb/snowflake-connector-python),
            # snowflake-connect-python requires at least Python 3.7
            'snowflake-connector-python[pandas]; python_version>="3.7"',
        ]
    },
    zip_safe                = False,
    include_package_data    = True,
    python_requires         = ">=3.6",
    platforms               = "Linux, Mac OS X, Windows",
    keywords                = ["Jupyter", "JupyterLab", "JupyterLab3"],
    classifiers             = [
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Framework :: Jupyter",
    ],
)

if __name__ == '__main__':
    setup(**setup_args)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.


from typing import Any, Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import (
    column_header_list_to_transpiled_code, column_header_to_transpiled_code)
from mitosheet.types import ColumnID, ColumnHeader

LOOKUP = 'lookup'
UNIQUE_IN_LEFT = 'unique in left'
UNIQUE_IN_RIGHT = 'unique in right'


class MergeCodeChunk(CodeChunk):


    def __init__(self, prev_state: State, post_state: State, how: str, sheet_index_one: int, sheet_index_two: int, merge_key_column_ids: List[List[ColumnID]], selected_column_ids_one: List[ColumnID], selected_column_ids_two: List[ColumnID]):
        super().__init__(prev_state, post_state)
        self.how: str = how 
        self.sheet_index_one: int = sheet_index_one 
        self.sheet_index_two: int = sheet_index_two 
        self.merge_key_column_ids: List[List[ColumnID]] = merge_key_column_ids 
        self.selected_column_ids_one: List[ColumnID] = selected_column_ids_one 
        self.selected_column_ids_two: List[ColumnID] = selected_column_ids_two

        self.df_one_name = self.post_state.df_names[self.sheet_index_one]
        self.df_two_name = self.post_state.df_names[self.sheet_index_two]
        self.df_new_name = self.post_state.df_names[len(self.post_state.dfs) - 1]

    def get_display_name(self) -> str:
        return 'Merged'
    
    def get_description_comment(self) -> str:
        return f'Merged {self.df_one_name} and {self.df_two_name} into {self.df_new_name}'

    def get_code(self) -> Tuple[List[str], List[str]]:
        merge_keys_one: List[ColumnHeader] = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index_one, list(map(lambda x: x[0], self.merge_key_column_ids)))
        merge_keys_two: List[ColumnHeader] = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index_two, list(map(lambda x: x[1], self.merge_key_column_ids)))

        selected_column_headers_one: List[ColumnHeader] = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index_one, self.selected_column_ids_one)
        selected_column_headers_two: List[ColumnHeader] = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index_two, self.selected_column_ids_two)

        if len(merge_keys_one) == 0 and len(merge_keys_two) == 0:
            return [f'{self.df_new_name} = pd.DataFrame()'], ['import pandas as pd']

        # Now, we build the merge code 
        merge_code = []
        if self.how == 'lookup':
            # If the mege is a lookup, then we add the drop duplicates code
            temp_df_name = 'temp_df'
            merge_code.append(f'{temp_df_name} = {self.df_two_name}.drop_duplicates(subset={column_header_list_to_transpiled_code(merge_keys_two)}) # Remove duplicates so lookup merge only returns first match')
            how_to_use = 'left'
        else:
            temp_df_name = self.df_two_name
            how_to_use = self.how

        # If we are only taking some columns, write the code to drop the ones we don't need!
        deleted_columns_one = set(self.post_state.dfs[self.sheet_index_one].keys()).difference(set(selected_column_headers_one).union(set(merge_keys_one)))
        deleted_columns_two = set(self.post_state.dfs[self.sheet_index_two].keys()).difference(set(selected_column_headers_two).union(set(merge_keys_two)))

        if len(deleted_columns_one) > 0:
            deleted_transpiled_column_header_one_list = column_header_list_to_transpiled_code(deleted_columns_one)
            merge_code.append(
                f'{self.df_one_name}_tmp = {self.df_one_name}.drop({deleted_transpiled_column_header_one_list}, axis=1)'
            )
        if len(deleted_columns_two) > 0:
            deleted_transpiled_column_header_two_list = column_header_list_to_transpiled_code(deleted_columns_two)
            merge_code.append(
                f'{self.df_two_name}_tmp = {temp_df_name}.drop({deleted_transpiled_column_header_two_list}, axis=1)'
            )

        # If we drop columns, we merge the new dataframes
        df_one_to_merge = self.df_one_name if len(deleted_columns_one) == 0 else f'{self.df_one_name}_tmp'
        df_two_to_merge = temp_df_name if len(deleted_columns_two) == 0 else f'{self.df_two_name}_tmp'

        # We insist column names are unique in dataframes, so we default the suffixes to be the dataframe names
        suffix_one = self.df_one_name
        suffix_two = self.df_two_name if self.df_two_name != self.df_one_name else f'{self.df_two_name}_2'

        # Finially append the merge
        if self.how == UNIQUE_IN_LEFT:
            merge_code.append(
                f'{df_two_to_merge}_tmp = {df_two_to_merge}.drop_duplicates(subset={column_header_list_to_transpiled_code(merge_keys_two)})'
            )
            merge_code.append(
                f'bool_index_array = {df_one_to_merge}.merge({df_two_to_merge}_tmp, left_on={column_header_list_to_transpiled_code(merge_keys_one)}, right_on={column_header_list_to_transpiled_code(merge_keys_two)}, how=\'left\', indicator=True)[\'_merge\'] == \'left_only\''
            )
            merge_code.append(
                f'{self.df_new_name} = {df_one_to_merge}.copy(deep=True)[bool_index_array][{column_header_list_to_transpiled_code(selected_column_headers_one)}].reset_index(drop=True)'
            )
        elif self.how == UNIQUE_IN_RIGHT:
            merge_code.append(
                f'{df_one_to_merge}_tmp = {df_one_to_merge}.drop_duplicates(subset={column_header_list_to_transpiled_code(merge_keys_one)})'
            )
            merge_code.append(
                f'bool_index_array = {df_one_to_merge}_tmp.merge({df_two_to_merge}, left_on={column_header_list_to_transpiled_code(merge_keys_one)}, right_on={column_header_list_to_transpiled_code(merge_keys_two)}, how=\'right\', indicator=True)[\'_merge\'] == \'right_only\''
            )
            merge_code.append(
                f'{self.df_new_name} = {df_two_to_merge}.copy(deep=True)[bool_index_array][{column_header_list_to_transpiled_code(selected_column_headers_two)}].reset_index(drop=True)'
            )
        else:      
            merge_code.append(
                f'{self.df_new_name} = {df_one_to_merge}.merge({df_two_to_merge}, left_on={column_header_list_to_transpiled_code(merge_keys_one)}, right_on={column_header_list_to_transpiled_code(merge_keys_two)}, how=\'{how_to_use}\', suffixes=[\'_{suffix_one}\', \'_{suffix_two}\'])'
            )

        # And then return it
        return merge_code, []

    def get_created_sheet_indexes(self) -> List[int]:
        return [len(self.post_state.dfs) - 1]



from typing import Any, Callable


try:
    import os
    from ipykernel import get_connection_file
    from ipykernel.comm import Comm

    def get_current_kernel_id() -> str:
        """
        Return the url of the current jupyter notebook server.
        
        Adapted from: https://github.com/jupyter/notebook/issues/3156
        """
        # path/to/file/kernel-c13462b6-c6a2-4e17-b891-7f7847204df9.json
        connection_file = get_connection_file() # type: ignore
        # kernel-c13462b6-c6a2-4e17-b891-7f7847204df9.json
        file_name: str = os.path.basename(connection_file) 
        return file_name[len('kernel-'):-len('.json')] 
    
except ImportError:

    # If we're running in jupyterlite, we don't have access to the ipykernel module.
    # So we just return a dummy kernel id. And on the frontend, if we get this kernel
    # id, we just do our best to find the right kernel to establish a comm with
    def get_current_kernel_id() -> str:
        return 'kernel-00000000-0000-0000-0000-000000000000'
    
    class Comm: # type: ignore
        def send(
            self,
            data: Any,
            metadata: Any,
            buffers: Any
        ) -> None:
            pass

        def on_msg(
            self,
            callback: Callable
        ) -> None:
            pass



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.promote_row_to_header_code_chunk import PromoteRowToHeaderCodeChunk
from mitosheet.errors import make_invalid_promote_row_to_header

from mitosheet.state import State
from mitosheet.step_performers.column_steps.rename_column import rename_column_headers_in_state
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import ColumnHeader
from mitosheet.utils import MAX_ROWS, convert_df_to_parsed_json
from mitosheet.public.v3 import deduplicate_column_headers

def get_should_deduplicate_column_headers(column_headers: List[ColumnHeader]) -> bool:
    """
    Check if the column headers should be deduplicated, taking special care to 
    handle nans.
    """
    return len(set(column_headers)) != len(column_headers) or \
        sum(isinstance(ch, float) and pd.isna(ch) for ch in column_headers) >= 2
        

class PromoteRowToHeaderStepPerformer(StepPerformer):
    """
    Allows you to promote a row to a header (and deletes that row in the process).
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'promote_row_to_header'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        index: Any = get_param(params, 'index')

        # We make a new state to modify it
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])

        pandas_processing_time = 0.0

        df = post_state.dfs[sheet_index]

        new_headers = df.loc[index].tolist()
        new_headers = deduplicate_column_headers(new_headers)

        # Then, get all the column headers ids, before renamding them, so we don't have ordering bugs
        id_to_new_heaeder = {post_state.column_ids.get_column_id_by_header(sheet_index, old): new for old, new in zip(df.columns, new_headers)}
        # Then, update them in the state
        for column_id, new_column_header in id_to_new_heaeder.items():
             post_state.column_ids.set_column_header(sheet_index, column_id, new_column_header)

        # And then update the columns
        pandas_start_time_drop = perf_counter()
        df.columns = new_headers
        pandas_processing_time_drop = perf_counter() - pandas_start_time_drop


        post_state.dfs[sheet_index].drop(labels=[index], inplace=True)

        # We make sure that this making of headers will work, and not cause issues 
        # later while trying to convert to json. We throw an error if this causes
        # errors. See this bug: https://github.com/mito-ds/monorepo/issues/267
        try:
            convert_df_to_parsed_json(post_state.dfs[sheet_index].head(1))
        except:
            raise make_invalid_promote_row_to_header()

        return post_state, {
            'pandas_processing_time': pandas_processing_time + pandas_processing_time_drop,
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            PromoteRowToHeaderCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'), 
                get_param(params, 'index'),
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
The main entry point for the mitosheet package, this command
line interface allows you to set some toggles in the user.json
"""
import sys
from mitosheet.experiments.experiment_utils import set_experiment

from mitosheet.user.db import set_user_field
from mitosheet.user import initialize_user
from mitosheet.user.schemas import UJ_MITOSHEET_PRO, UJ_MITOSHEET_TELEMETRY, UJ_RECEIVED_CHECKLISTS, UJ_RECEIVED_TOURS, UJ_USER_EMAIL
from mitosheet.startup.startup_utils import create_startup_file, remove_startup_file
from mitosheet.user.utils import check_pro_acccess_code


def main() -> None:
    """
    Currently, the main usage of this function is:
    python -m mitosheet turnofflogging
    python -m mitosheet turnonlogging
    python -m mitosheet turnoffpro
    python -m mitosheet turnoffdataframebutton
    python -m mitosheet turnondataframebutton

    We also have 
    python -m mitosheet clearemail
    """
    # Make sure the user is initalized first, but do not identify
    # then, in case they are turning off logging
    initialize_user(call_identify=False)

    # Then, if we are being told to turn off logging, turn off logging
    if len(sys.argv) > 1:
        if sys.argv[-1] == 'turnofflogging':
            print("Turning off all logging")
            set_user_field(UJ_MITOSHEET_TELEMETRY, False)
            print("Logging turned off!")
        if sys.argv[-1] == 'turnonlogging':
            print("Turning on all logging")
            set_user_field(UJ_MITOSHEET_TELEMETRY, True)
            print("Logging turned on!")
        if sys.argv[-1] == 'turnoffpro':
            print("Turning off Mitosheet pro")
            set_user_field(UJ_MITOSHEET_PRO, False)
            print("Mitosheet Pro is false!")
        if sys.argv[-2] == 'turnonpro':
            print("Turning on Mitosheet pro", sys.argv[-1])
            if check_pro_acccess_code(sys.argv[-1].strip()):
                set_user_field(UJ_MITOSHEET_PRO, True)
                set_user_field(UJ_MITOSHEET_TELEMETRY, False)
                print("Mitosheet Pro is now active!")
            else:
                print("Invalid access code...")
        if sys.argv[-1] == 'clearemail':
            print("Clearing email")
            set_user_field(UJ_USER_EMAIL, '')
            print("Email cleared")
        if sys.argv[-1] == 'cleartours':
            print("Clearing tours")
            set_user_field(UJ_RECEIVED_TOURS, [])
        if sys.argv[-1] == 'clearchecklists':
            print("Clearing checklists")
            set_user_field(UJ_RECEIVED_CHECKLISTS, {})
            print("Checklists cleared")
        if sys.argv[-1] == 'turnoffdataframebutton':
            print("Turning off the 'View in Mito' dataframe button")
            remove_startup_file()
            print("Turned off the 'View in Mito' dataframe button\nThe next time you launch Jupyter Lab, the button will no longer be visible")
        if sys.argv[-1] == 'turnondataframebutton':
            print("Turning on the 'View in Mito' dataframe button")
            create_startup_file()
            print("Turned on the 'View in Mito' dataframe button\nThe next time you launch Jupyter Lab, the button will be visible")
        if len(sys.argv) > 2:
            if sys.argv[-3] == 'setexperiment':
                print("Setting experiment", sys.argv[-2], "to variant", sys.argv[-1])
                set_experiment(sys.argv[-2], sys.argv[-1])
                print("Experiment set")
        
if __name__ == '__main__':
    main()


from datetime import datetime, timedelta
from typing import Optional, Union

import numpy as np


def cast_string_to_bool(
        s: str,
    ) -> Optional[bool]:

    string_to_bool_conversion_dict = {
        '1': True,
        '1.0': True,
        1: True,
        1.0: True,
        'TRUE': True,
        'True': True, 
        'true': True,
        'T': True,
        't': True,
        'Y': True,
        'y': True,
        'Yes': True,
        'yes': True,
        #########################
        '0': False,
        '0.0': False,
        0: False,
        0.0: False,
        'FALSE': False,
        'False': False,
        'false': False,
        'F': False,
        'f': False,
        'N': False,
        'n': False,
        'No': False,
        'no': False, 
        'none': False,
        'None': False
    }

    if s in string_to_bool_conversion_dict:
        return string_to_bool_conversion_dict[s]
    else:
        return None # TODO: maybe we should default to False


def cast_to_bool(unknown: Union[str, int, float, bool, datetime, timedelta]) -> Optional[bool]:
    if isinstance(unknown, str):
        return cast_string_to_bool(unknown)
    elif isinstance(unknown, int):
        return bool(unknown)
    elif isinstance(unknown, float):
        # We cast NaN's to false
        if np.isnan(unknown):
            return False

        return bool(unknown)
    elif isinstance(unknown, bool):
        return unknown

    return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

import os
from mitosheet.enterprise.mito_config import (
    DEFAULT_MITO_CONFIG_SUPPORT_EMAIL, 
    MEC_VERSION_KEYS,
    MITO_CONFIG_ANALYTICS_URL, 
    MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL, 
    MITO_CONFIG_CODE_SNIPPETS, 
    MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL, 
    MITO_CONFIG_CODE_SNIPPETS_URL, 
    MITO_CONFIG_CODE_SNIPPETS_VERSION,
    MITO_CONFIG_DISABLE_TOURS,
    MITO_CONFIG_LLM_URL, 
    MITO_CONFIG_SUPPORT_EMAIL, 
    MITO_CONFIG_VERSION, 
    MITO_CONFIG_SUPPORT_EMAIL, 
    MITO_CONFIG_VERSION, 
    MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT,
    MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT,
    MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION,
    MITO_CONFIG_FEATURE_TELEMETRY,
    MITO_CONFIG_PRO,
    MitoConfig
)


def delete_env_var_if_exists(env_var: str) -> None: 
    """
    Deletes the environment variable only if it exists to avoid errors. Helpful for testing.
    """
    if os.environ.get(env_var) is not None:
        del os.environ[env_var]

def delete_all_mito_config_environment_variables() -> None:
    """
    Deletes all of the MITO CONFIG environment variables that are used in the MEC
    """
    lists_of_vars = MEC_VERSION_KEYS.values()
    for list_of_vars in lists_of_vars:
        for var in list_of_vars:
            delete_env_var_if_exists(var)

def test_keys_did_not_change():
    # We must not change these keys so we can still read old 
    # mito configs that users have not upgraded to the most 
    # recent config version. If we don't preserve these keys, 
    # we won't know which environment variables to read.
    assert MITO_CONFIG_VERSION == 'MITO_CONFIG_VERSION'
    assert MITO_CONFIG_SUPPORT_EMAIL == 'MITO_CONFIG_SUPPORT_EMAIL'
    assert MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL == 'MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL'

def test_none_works():
    # Delete the environmnet variables so we can test the none condition
    delete_all_mito_config_environment_variables()

    mito_config = MitoConfig()
    mito_config_dict = mito_config.get_mito_config()
    assert mito_config_dict == {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: DEFAULT_MITO_CONFIG_SUPPORT_EMAIL,
        MITO_CONFIG_DISABLE_TOURS: False,
        MITO_CONFIG_CODE_SNIPPETS: None,
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: False,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: True,
        MITO_CONFIG_LLM_URL: None,
        MITO_CONFIG_ANALYTICS_URL: None,
        MITO_CONFIG_FEATURE_TELEMETRY: True,
        MITO_CONFIG_PRO: False
    }

def test_none_config_version_is_string():
    # Delete the environmnet variables so we can test the none condition
    delete_all_mito_config_environment_variables()

    mito_config = MitoConfig()
    mito_config_dict = mito_config.get_mito_config()
    assert isinstance(mito_config_dict[MITO_CONFIG_VERSION], str)

def test_version_2_works():
    # Set environment variables
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_SUPPORT_EMAIL] = "aaron@sagacollab.com"
    os.environ[MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL] = "jake@sagacollab.com"
    os.environ[MITO_CONFIG_CODE_SNIPPETS_URL] = "url"
    os.environ[MITO_CONFIG_DISABLE_TOURS] = "True"
    os.environ[MITO_CONFIG_CODE_SNIPPETS_VERSION] = "1"

    # Test reading environment variables works properly
    mito_config = MitoConfig()
    assert mito_config.get_mito_config() == {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: 'aaron@sagacollab.com',
        MITO_CONFIG_DISABLE_TOURS: True,
        MITO_CONFIG_CODE_SNIPPETS: {
            MITO_CONFIG_CODE_SNIPPETS_VERSION : '1',
            MITO_CONFIG_CODE_SNIPPETS_URL: 'url',
            MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL: "jake@sagacollab.com"
        },
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: False,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: True,
        MITO_CONFIG_LLM_URL: None,
        MITO_CONFIG_ANALYTICS_URL: None,
        MITO_CONFIG_FEATURE_TELEMETRY: True,
        MITO_CONFIG_PRO: False
    }

    # Delete the environmnet variables for the next test
    delete_all_mito_config_environment_variables()

def test_mito_config_update_version_1_to_2():
    # Set environment variables
    os.environ[MITO_CONFIG_VERSION] = "1"
    os.environ[MITO_CONFIG_SUPPORT_EMAIL] = "aaron@sagacollab.com"
    
    # Test reading environment variables works properly
    mito_config = MitoConfig()
    assert mito_config.get_mito_config() == {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: 'aaron@sagacollab.com',
        MITO_CONFIG_DISABLE_TOURS: False,
        MITO_CONFIG_CODE_SNIPPETS: None,
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: False,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: True,
        MITO_CONFIG_LLM_URL: None,
        MITO_CONFIG_ANALYTICS_URL: None,
        MITO_CONFIG_FEATURE_TELEMETRY: True,
        MITO_CONFIG_PRO: False
    }    

    # Delete the environmnet variables for the next test
    delete_all_mito_config_environment_variables()

def test_mito_config_enable_snowflake_import():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_SUPPORT_EMAIL] = "aaron@sagacollab.com"
    os.environ[MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT] = "True"
    
    mito_config = MitoConfig()
    assert mito_config.get_mito_config() == {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: 'aaron@sagacollab.com',
        MITO_CONFIG_DISABLE_TOURS: False,
        MITO_CONFIG_CODE_SNIPPETS: None,
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: True,
        MITO_CONFIG_LLM_URL: None,
        MITO_CONFIG_ANALYTICS_URL: None,
        MITO_CONFIG_FEATURE_TELEMETRY: True,
        MITO_CONFIG_PRO: False
    }    

    delete_all_mito_config_environment_variables()

def test_mito_config_dont_display_snowflake_import():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_SUPPORT_EMAIL] = "aaron@sagacollab.com"
    os.environ[MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT] = "True"
    os.environ[MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT] = "False"
    
    mito_config = MitoConfig()
    assert mito_config.get_mito_config() == {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: 'aaron@sagacollab.com',
        MITO_CONFIG_DISABLE_TOURS: False,
        MITO_CONFIG_CODE_SNIPPETS: None,
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: False,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: True,
        MITO_CONFIG_LLM_URL: None,
        MITO_CONFIG_ANALYTICS_URL: None,
        MITO_CONFIG_FEATURE_TELEMETRY: True,
        MITO_CONFIG_PRO: False
    }    

    delete_all_mito_config_environment_variables()

def test_mito_config_dont_display_ai_transform():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION] = "False"
    
    mito_config = MitoConfig()
    assert mito_config.get_mito_config() == {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: 'founders@sagacollab.com',
        MITO_CONFIG_DISABLE_TOURS: False,
        MITO_CONFIG_CODE_SNIPPETS: None,
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: False,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: False,
        MITO_CONFIG_LLM_URL: None,
        MITO_CONFIG_ANALYTICS_URL: None,
        MITO_CONFIG_FEATURE_TELEMETRY: True,
        MITO_CONFIG_PRO: False
    }    

    delete_all_mito_config_environment_variables()

def test_mit_config_disable_telemetry():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_FEATURE_TELEMETRY] = "False"

    from mitosheet.telemetry.telemetry_utils import telemetry_turned_on
    assert not telemetry_turned_on()


def test_mit_config_enable_pro_telemetry():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_PRO] = "True"

    from mitosheet.user import is_pro
    assert is_pro()




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Exports the transpile function, which takes the backend widget
container and generates transpiled Python code.
"""

from typing import Any, Dict, List
from mitosheet.array_utils import deduplicate_array
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.code_chunk_utils import get_code_chunks
from mitosheet.code_chunks.postprocessing import POSTPROCESSING_CODE_CHUNKS

from mitosheet.preprocessing import PREPROCESS_STEP_PERFORMERS
from mitosheet.transpiler.transpile_utils import convert_script_to_function
from mitosheet.types import StepsManagerType


IN_PREVIOUS_STEP_COMMENT = '# You\'re viewing a previous step. Click fast forward in the Mitosheet above to see the full analysis.'


IMPORT_STATEMENTS = {
    1: 'from mitosheet.public.v1 import *',
    2: 'from mitosheet.public.v2 import *',
    3: 'from mitosheet.public.v3 import *'
}

def transpile(
        steps_manager: StepsManagerType, 
        add_comments: bool=True,
        optimize: bool=True
    ) -> List[str]:
    """
    Transpiles the code from the current steps in the steps_manager, 
    displaying up to the checked out step.

    If add_comments, then adds descriptive comments using the step 
    describe functions. 
    """

    imports_code: List[str] = []
    code: List[str] = []

    # First, we transpile all the preprocessing steps
    for preprocess_step_performer in PREPROCESS_STEP_PERFORMERS:
        preprocess_code, preprocess_imports = preprocess_step_performer.transpile(
            steps_manager,
            steps_manager.preprocess_execution_data[preprocess_step_performer.preprocess_step_type()],
        )
        if len(preprocess_code) > 0:
            code.extend(preprocess_code)
            code.append('')

        imports_code.extend(preprocess_imports)

    # We only transpile up to the currently checked out step
    all_code_chunks: List[CodeChunk] = get_code_chunks(steps_manager.steps_including_skipped[:steps_manager.curr_step_idx + 1], optimize=optimize)

    # We also make sure to include all the post_processing code chunks, which are those
    # code chunks that are always at the end of the dataframe
    for postprocessing_code_chunk in POSTPROCESSING_CODE_CHUNKS:
        all_code_chunks.append(postprocessing_code_chunk(steps_manager.curr_step.initial_defined_state, steps_manager.curr_step.final_defined_state))

    for code_chunk in all_code_chunks:
        comment = '# ' + code_chunk.get_description_comment().strip().replace('\n', '\n# ')
        (gotten_code, code_chunk_imports) = code_chunk.get_code()

        # Make sure to not generate comments or code for steps with no code 
        if len(gotten_code) > 0:
            if add_comments:
                gotten_code.insert(0, comment)
            code.extend(gotten_code)

            # Then add a line of whitespace
            code.append('')

        imports_code.extend(code_chunk_imports)

    # If we have a historical step checked out, then we add a comment letting
    # the user know this is the case
    if steps_manager.curr_step_idx != len(steps_manager.steps_including_skipped) - 1:
        code.append(IN_PREVIOUS_STEP_COMMENT)

    # We then deduplicate the imports, keeping the same order as originally
    final_imports_code = deduplicate_array(imports_code)

    # We then add the import statement for the correct version of mitosheet
    if len(code) > 0:
        final_imports_code.insert(0, IMPORT_STATEMENTS[steps_manager.public_interface_version])

    # If we should transpile this as a function, we do so
    if steps_manager.code_options['as_function']:
        final_code = convert_script_to_function(
            steps_manager,
            final_imports_code,
            code,
            steps_manager.code_options['function_name'],
            steps_manager.code_options['function_params'],
            steps_manager.code_options['call_function']
        )
        return final_code

    # We add some whitespace if there are imports
    if len(final_imports_code) > 0:
        final_imports_code.append('')

    return final_imports_code + code



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.import_steps.simple_import_code_chunk import DEFAULT_DECIMAL
from mitosheet.state import State
from mitosheet.step_performers.utils import get_param
from mitosheet.transpiler.transpile_utils import column_header_to_transpiled_code


class ExcelImportCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, file_name: str, sheet_names: List[str], has_headers: bool, skiprows: int, decimal: str):
        super().__init__(prev_state, post_state)
        self.file_name = file_name
        self.sheet_names = sheet_names
        self.has_headers = has_headers
        self.skiprows = skiprows
        self.decimal = decimal

    def get_display_name(self) -> str:
        return 'Imported'
    
    def get_description_comment(self) -> str:
        return f'Imported {", ".join(self.sheet_names)} from {self.file_name}'

    def get_code(self) -> Tuple[List[str], List[str]]:
        read_excel_params = build_read_excel_params(
            self.sheet_names,
            self.has_headers,
            self.skiprows,
            self.decimal
        )

        tranpiled_file_name = column_header_to_transpiled_code(self.file_name)
        read_excel_line = f'sheet_df_dictonary = pd.read_excel(r{tranpiled_file_name}, engine=\'openpyxl\''
        for key, value in read_excel_params.items():
            # We use this slighly misnamed function to make sure values get transpiled right
            read_excel_line += f", {key}={column_header_to_transpiled_code(value)}"
        read_excel_line += ')'

        df_definitions = []
        for index, sheet_name in enumerate(self.sheet_names):
            adjusted_index = len(self.post_state.df_names) - len(self.sheet_names) + index
            df_definitions.append(
                f'{self.post_state.df_names[adjusted_index]} = sheet_df_dictonary[\'{sheet_name}\']'
            )

        return [
            read_excel_line
        ] + df_definitions, ['import pandas as pd']

    def get_created_sheet_indexes(self) -> List[int]:
        return [i for i in range(len(self.post_state.dfs) - len(self.sheet_names), len(self.post_state.dfs))]
    
    def get_parameterizable_params(self) -> List[Tuple[str, str, str]]:
        return [(f'r{column_header_to_transpiled_code(self.file_name)}', 'file_name', 'Excel import file path')]

    
def build_read_excel_params(
    sheet_names: List[str],
    has_headers: bool,
    skiprows: int,
    decimal: str
) -> Dict[str, Any]:
    read_excel_params = {
        'sheet_name': sheet_names,
        'skiprows': skiprows,
    }

    # Get rid of the headers if it doesn't have them
    if not has_headers:
        read_excel_params['header'] = None

    if decimal is not None and decimal is not DEFAULT_DECIMAL: 
        read_excel_params['decimal'] = decimal

    return read_excel_params


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import deepcopy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.column_steps.split_text_to_columns_code_chunk import SplitTextToColumnsCodeChunk, get_split_param_dict
from mitosheet.is_type_utils import is_datetime_dtype, is_timedelta_dtype
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import add_columns_to_df, get_param
from mitosheet.types import ColumnHeader, ColumnID


class SplitTextToColumnsStepPerformer(StepPerformer):
    """""
    A split_text_to_columns step, which allows you to separate 
    a single column into multiple columns by specifying a delimiter
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'split_text_to_columns'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_id: ColumnID = get_param(params, 'column_id')
        delimiters: List[str] = get_param(params, 'delimiters')
        new_column_header_suffix: str = get_param(params, 'new_column_header_suffix')

        column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)
        column_idx = prev_state.dfs[sheet_index].columns.tolist().index(column_header)
            
        # Create a new post state
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])
        final_df = post_state.dfs[sheet_index]
        delimiter_string = '|'.join(delimiters)
        # Actually execute the column reordering
        pandas_start_time = perf_counter() 
            
        split_param_dict = get_split_param_dict()

        # Create the dataframe of new columns. We do this first, so that we know how many columns get created.
        if is_datetime_dtype(str(post_state.dfs[sheet_index][column_header].dtype)):
            new_columns_df = final_df[column_header].dt.strftime('%Y-%m-%d %X').str.split(delimiter_string, **split_param_dict)
        elif is_timedelta_dtype(str(post_state.dfs[sheet_index][column_header].dtype)):
            new_columns_df = final_df[column_header].apply(lambda x: str(x)).str.split(delimiter_string, **split_param_dict)
        else:
            new_columns_df = final_df[column_header].astype('str').str.split(delimiter_string, **split_param_dict)

        # Create the new column headers and ensure they are unique
        # Note: We create the new_column_header_suffix on the frontend so that it is saved in the step parameters, which allows us
        # to replay the analysis and generate the same columns. 
        new_column_headers: List[ColumnHeader] = [f'{column_header}-split-{idx}-{new_column_header_suffix}' for column, idx in enumerate(new_columns_df)]
        final_df = add_columns_to_df(final_df, new_columns_df, new_column_headers, column_idx)
        pandas_processing_time = perf_counter() - pandas_start_time

        # Update column state variables
        post_state.add_columns_to_state(sheet_index, new_column_headers)

        post_state.dfs[sheet_index] = final_df

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            # Save the new_column_headers so that the code chunk doesn't need to call .split to figure out how many columns were created
            'new_column_headers': new_column_headers,
            'result': {
                'num_cols_created': len(new_column_headers)
            }
        }

        
    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            SplitTextToColumnsCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_id'),
                get_param(params, 'delimiters'),
                execution_data.get('new_column_headers', []) if execution_data is not None else []
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}



from typing import Any, Optional, Dict

from openpyxl.styles import Font, PatternFill
from openpyxl.styles import NamedStyle
from openpyxl.formatting.rule import FormulaRule
from openpyxl.worksheet.worksheet import Worksheet
from pandas import DataFrame, ExcelWriter

from mitosheet.excel_utils import get_column_from_column_index
from mitosheet.types import (
    FC_BOOLEAN_IS_FALSE, FC_BOOLEAN_IS_TRUE, FC_DATETIME_EXACTLY,
    FC_DATETIME_GREATER, FC_DATETIME_GREATER_THAN_OR_EQUAL, FC_DATETIME_LESS,
    FC_DATETIME_LESS_THAN_OR_EQUAL, FC_DATETIME_NOT_EXACTLY, FC_EMPTY,
    FC_LEAST_FREQUENT, FC_MOST_FREQUENT, FC_NOT_EMPTY, FC_NUMBER_EXACTLY,
    FC_NUMBER_GREATER, FC_NUMBER_GREATER_THAN_OR_EQUAL, FC_NUMBER_HIGHEST,
    FC_NUMBER_LESS, FC_NUMBER_LESS_THAN_OR_EQUAL, FC_NUMBER_LOWEST,
    FC_NUMBER_NOT_EXACTLY, FC_STRING_CONTAINS, FC_STRING_DOES_NOT_CONTAIN,
    FC_STRING_ENDS_WITH, FC_STRING_EXACTLY, FC_STRING_NOT_EXACTLY,
    FC_STRING_STARTS_WITH, FC_STRING_CONTAINS_CASE_INSENSITIVE)


# Object to map the conditional formatting operators to the excel formulas
CONDITION_TO_COMPARISON_FORMULA: Dict[str, str] = {
    FC_NUMBER_GREATER: '>',
    FC_NUMBER_LESS: '<',
    FC_NUMBER_EXACTLY: '=',
    FC_NUMBER_NOT_EXACTLY: '<>',
    FC_NUMBER_GREATER_THAN_OR_EQUAL: '>=',
    FC_NUMBER_LESS_THAN_OR_EQUAL: '<=',
    FC_DATETIME_EXACTLY: '=',
    FC_DATETIME_NOT_EXACTLY: '<>',
    FC_DATETIME_GREATER: '>',
    FC_DATETIME_GREATER_THAN_OR_EQUAL: '>=',
    FC_DATETIME_LESS: '<',
    FC_DATETIME_LESS_THAN_OR_EQUAL: '<=',
}

def get_conditional_format_rule(
    filter_condition: str,
    fill: Optional[PatternFill],
    font: Optional[Font],
    filter_value: str,
    cell_range: str,
    column: str
) -> Optional[FormulaRule]:
    # Update the formulas for the string operators
    comparison = CONDITION_TO_COMPARISON_FORMULA.get(filter_condition)
    # If comparing dates, we need to use the DATEVALUE function
    if comparison is not None and 'datetime' in filter_condition:
        formula = [f'{cell_range}{comparison}DATEVALUE("{filter_value}")']
    elif comparison is not None:
        formula = [f'{cell_range}{comparison}{filter_value}']
    elif filter_condition == FC_STRING_CONTAINS:
        formula = [f'NOT(ISERROR(FIND("{filter_value}",{cell_range})))']
    elif filter_condition == FC_STRING_CONTAINS_CASE_INSENSITIVE:
        formula = [f'NOT(ISERROR(SEARCH("{filter_value}",{cell_range})))']
    elif filter_condition == FC_STRING_DOES_NOT_CONTAIN:
        formula = [f'ISERROR(SEARCH("{filter_value}",{cell_range}))']
    elif filter_condition == FC_STRING_STARTS_WITH:
        formula = [f'LEFT({cell_range},LEN("{filter_value}"))="{filter_value}"']
    elif filter_condition == FC_STRING_ENDS_WITH:
        formula = [f'RIGHT({cell_range},LEN("{filter_value}"))="{filter_value}"']
    elif filter_condition == FC_BOOLEAN_IS_TRUE:
        formula = [cell_range]
    elif filter_condition == FC_BOOLEAN_IS_FALSE:
        formula = [f'NOT({cell_range})']
    elif filter_condition == FC_STRING_EXACTLY:
        formula = [f'EXACT({cell_range},"{filter_value}")']
    elif filter_condition == FC_STRING_NOT_EXACTLY:
        formula = [f'NOT(EXACT({cell_range},"{filter_value}"))']
    elif filter_condition == FC_EMPTY:
        formula = [f'ISBLANK({cell_range})']
    elif filter_condition == FC_NOT_EMPTY:
        formula = [f'NOT(ISBLANK({cell_range}))']
    elif filter_condition == FC_NUMBER_HIGHEST:
        formula = [f'{column}2>=LARGE({column}:{column},{filter_value})']
    elif filter_condition == FC_NUMBER_LOWEST:
        formula = [f'{column}2<=SMALL({column}:{column},{filter_value})']
    else: return None
    return FormulaRule(fill=fill, font=font, formula=formula)

def add_conditional_formats(
    conditional_formats: list,
    sheet: Worksheet,
    df: DataFrame
) -> None:
    for conditional_format in conditional_formats:
        for filter in conditional_format.get('filters', []):
            # Create the conditional formatting color objects
            cond_fill = None
            cond_font = None
            if conditional_format.get('background_color') is not None:
                cond_fill = PatternFill(start_color=conditional_format['background_color'][1:], end_color=conditional_format['background_color'][1:], fill_type='solid')
            if conditional_format.get('font_color') is not None:
                cond_font = Font(color=conditional_format['font_color'][1:])
            if cond_fill is None and cond_font is None:
                continue
            
            # Add the conditional formatting rule to the sheet
            for column_header in conditional_format['columns']:
                column_index = df.columns.tolist().index(column_header)
                column = get_column_from_column_index(column_index)
                cell_range = f'{column}2:{column}{sheet.max_row}'
                column_conditional_rule = get_conditional_format_rule(
                    filter_condition=filter['condition'],
                    fill=cond_fill,
                    font=cond_font,
                    filter_value=f"{filter['value']}",
                    cell_range=cell_range,
                    column=column
                )
                if column_conditional_rule is not None:
                    sheet.conditional_formatting.add(cell_range, column_conditional_rule)


def add_number_formatting(
    number_formats: Optional[Dict[str, str]],
    sheet: Worksheet,
    df: DataFrame
) -> None:
    if number_formats is None:
        return
    for column_header, number_format in number_formats.items():
        column_index = df.columns.tolist().index(column_header)
        column = get_column_from_column_index(column_index)
        cell_range = f'{column}2:{column}{sheet.max_row}'
        for cell in sheet[cell_range]:
            cell[0].number_format = number_format


def add_formatting_to_excel_sheet(
        writer: ExcelWriter,
        sheet_name: str,
        df: DataFrame,
        header_background_color: Optional[str]=None,
        header_font_color: Optional[str]=None,
        even_background_color: Optional[str]=None,
        even_font_color: Optional[str]=None,
        odd_background_color: Optional[str]=None,
        odd_font_color: Optional[str]=None,
        conditional_formats: Optional[list]=None,
        number_formats: Optional[Dict[str, Any]]=None
    ) -> None:
        """
        Adds formatting to the sheet_name, based on the formatting the user
        currently has applied in the frontend. 

        NOTE: this is a Mito Pro feature.
        """
        workbook = writer.book
        sheet = workbook.get_sheet_by_name(sheet_name)
        
        # Add formatting to the header row   
        header_name = f"{sheet_name}_Header"
        header_format = NamedStyle(name=header_name)
        if header_font_color:
            # Remove the # from the color
            header_format.font = Font(color=header_font_color[1:])
        if header_background_color:
            # Remove the # from the color
            header_format.fill = PatternFill(start_color=header_background_color[1:], end_color=header_background_color[1:], fill_type="solid")
        
        # Only add format if there is a header color or background color
        has_header_formatting = header_background_color or header_font_color 
        if has_header_formatting:
            # Add named styles for the header rows to improve performance
            workbook.add_named_style(header_format)

            # Write the formatting to the sheet
            for col in range(1, sheet.max_column + 1):
                sheet.cell(row=1, column=col).style = header_name

        # Add formatting to the rows
        even_name = f"{sheet_name}_Even"
        odd_name = f"{sheet_name}_Odd"
        even_format = NamedStyle(name=even_name)
        odd_format = NamedStyle(name=odd_name)

        # Remove the # from the colors and define the formatting objects
        if even_background_color:
            even_format.fill = PatternFill(start_color=even_background_color[1:], end_color=even_background_color[1:], fill_type="solid")
        if even_font_color:
            even_format.font = Font(color=even_font_color[1:])
        if odd_background_color:
            odd_format.fill = PatternFill(start_color=odd_background_color[1:], end_color=odd_background_color[1:], fill_type="solid")
        if odd_font_color:
            odd_format.font = Font(color=odd_font_color[1:])
        
        # Only add format if there is a background color or font color
        has_row_formatting = even_background_color or even_font_color or odd_background_color or odd_font_color
        if has_row_formatting:
            workbook.add_named_style(even_format)
            workbook.add_named_style(odd_format)

            for row in range(2, sheet.max_row + 1):
                for col in range(1, sheet.max_column + 1):
                    if row % 2 == 0:
                        sheet.cell(row=row, column=col).style = even_name
                    else:
                        sheet.cell(row=row, column=col).style = odd_name

        # Add conditional formatting
        if conditional_formats is not None:
            add_conditional_formats(conditional_formats, sheet, df)

        if number_formats is not None:
            add_number_formatting(number_formats, sheet, df)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
Contains some types that are useful in the Mitosheet package. 

We use type aliases to make many parts of the codebase more
explicit and clear, and make sure to test the types in our 
continous integration
"""

import pandas as pd
import numpy as np
from typing import TYPE_CHECKING, Dict, List, Optional, Union, Tuple, Any

GraphID = str
ColumnID = str

# A column header is either a primative type
PrimativeColumnHeader = Union[int, float, bool, str, Optional[str]]
MultiLevelColumnHeader = Union[Tuple[PrimativeColumnHeader, ...], List[PrimativeColumnHeader]]
# To a tuple of primative types (TODO: does this nest further?).
ColumnHeader = Union[PrimativeColumnHeader, MultiLevelColumnHeader]
IndexLabel = Any

# To resolve circular dependencies, we create a StepsManagerType here
if TYPE_CHECKING:
    from mitosheet.steps_manager import StepsManager
    StepsManagerType = StepsManager
    from mitosheet.mito_backend import MitoBackend
    MitoWidgetType = MitoBackend
    from mitosheet.state import State
    StateType = State
else:
    StepsManagerType = Any
    MitoWidgetType = Any
    StateType = Any

IndexType = Union[str, int, bool, float]


DataframeFormat = Dict[str, Any]
ColumnFormat = Dict[str, Any]


ConditionalFormatUUID = str

ConditionalFormat = Dict[str, Any] 
"""
ConditionalFormat: {
    format_uuid: string, // Should be a random string!
    columnIDs: ColumnID[],
    filters: FilterType[],
    invalidFilterColumnIDs: ColumnID[]
    color: string | undefined
    backgroundColor: string | undefined
}
"""

ConditionalFormattingInvalidResults = Dict[ConditionalFormatUUID, List[ColumnID]]
ConditionalFormattingCellResults = Dict[ColumnID, Dict[IndexType, Dict[str, Optional[str]]]]

ConditionalFormattingResult = Dict[str, Union[
        ConditionalFormattingInvalidResults, # A list of the invalid columns for a specific filter
        ConditionalFormattingCellResults # The actual formatting results
    ]
]

PivotColumnTransformation = str

RowOffset = int
ParserMatchSubstringRange = Tuple[int, int] # start, end

# If the user does not have the snowflake.connector python package installed,
# we take extra care to make sure that our mypy typing will still pass even though
# that code is not accessible to the user.
try:
    from snowflake.connector import SnowflakeConnection 
except ImportError:
    # This is prety hacky and causes the mypy tests to fail if we get 
    # rid of the #type: ignore, but I'm unable to create a type based on the 
    # SnowflakeConnection connection otherwise because its conditional 
    # on whether the package is installed or not.
    SnowflakeConnection = Any #type: ignore

MitoSafeSnowflakeConnection = Optional[SnowflakeConnection]

FORMULA_ENTIRE_COLUMN_TYPE = 'entire_column'
FORMULA_SPECIFIC_INDEX_LABELS_TYPE = 'specific_index_labels'

# The constants used in the filter step itself as filter conditions
# NOTE: these must be unique (e.g. no repeating names for different types)
FC_EMPTY = "empty"
FC_NOT_EMPTY = "not_empty"
FC_LEAST_FREQUENT = "least_frequent"
FC_MOST_FREQUENT = "most_frequent"

FC_BOOLEAN_IS_TRUE = "boolean_is_true"
FC_BOOLEAN_IS_FALSE = "boolean_is_false"

FC_STRING_CONTAINS = "contains"
FC_STRING_DOES_NOT_CONTAIN = "string_does_not_contain"
FC_STRING_EXACTLY = "string_exactly"
FC_STRING_NOT_EXACTLY = "string_not_exactly"
FC_STRING_STARTS_WITH = "string_starts_with"
FC_STRING_ENDS_WITH = "string_ends_with"
FC_STRING_CONTAINS_CASE_INSENSITIVE = "string_contains_case_insensitive"

FC_NUMBER_EXACTLY = "number_exactly"
FC_NUMBER_NOT_EXACTLY = "number_not_exactly"
FC_NUMBER_GREATER = "greater"
FC_NUMBER_GREATER_THAN_OR_EQUAL = "greater_than_or_equal"
FC_NUMBER_LESS = "less"
FC_NUMBER_LESS_THAN_OR_EQUAL = "less_than_or_equal"
FC_NUMBER_LOWEST = 'number_lowest'
FC_NUMBER_HIGHEST = 'number_highest'

FC_DATETIME_EXACTLY = "datetime_exactly"
FC_DATETIME_NOT_EXACTLY = "datetime_not_exactly"
FC_DATETIME_GREATER = "datetime_greater"
FC_DATETIME_GREATER_THAN_OR_EQUAL = "datetime_greater_than_or_equal"
FC_DATETIME_LESS = "datetime_less"
FC_DATETIME_LESS_THAN_OR_EQUAL = "datetime_less_than_or_equal"

ParamName = str
ParamValue = str

import sys
if sys.version_info[:3] > (3, 8, 0):
    from typing import TypedDict, Literal


    BooleanFilterCondition = Literal[
        'boolean_is_true',
        'boolean_is_false'
    ]
    StringFilterCondition = Literal[
        'contains',
        'string_does_not_contain',
        'string_exactly',
        'string_not_exactly',
        'string_starts_with',
        'string_ends_with'
    ]
    NumberFilterCondition = Literal[
        'number_exactly',
        'number_not_exactly',
        'greater',
        'greater_than_or_equal',
        'less',
        'less_than_or_equal',
        'number_lowest',
        'number_highest',
    ]
    DatetimeFilterCondition = Literal[
        'datetime_exactly',
        'datetime_not_exactly',
        'datetime_greater',
        'datetime_greater_than_or_equal',
        'datetime_less',
        'datetime_less_than_or_equal'
    ]
    SharedFilterCondition = Literal[
        'empty',
        'not_empty',
        'most_frequent',
        'least_frequent'
    ]
    OperatorType = Literal[
        'And', 
        'Or'
    ]
    
    PrimitiveTypeName = Literal[
        'str', 
        'int', 
        'float', 
        'number',
        'bool', 
        'datetime', 
        'timedelta'
    ]

    class Filter(TypedDict):
        condition: Union[BooleanFilterCondition, StringFilterCondition, NumberFilterCondition, DatetimeFilterCondition, SharedFilterCondition]
        value: Union[str, float, int]

    class FilterGroup(TypedDict):
        # NOTE: this is a recursive type. The filter group can contain a filter group
        filters: List[Union[Filter, "FilterGroup"]] #type: ignore
        operator: OperatorType #type:ignore

    class ColumnIDWithFilter(TypedDict):
        column_id: ColumnID
        filter: Filter

    class ColumnIDWithFilterGroup(TypedDict):
        column_id: ColumnID
        filter: FilterGroup

    class ColumnHeaderWithFilter(TypedDict):
        column_header: ColumnHeader
        filter: Filter

    class ColumnIDWithPivotTransform(TypedDict):
        column_id: ColumnID
        transformation: PivotColumnTransformation

    class ColumnHeaderWithPivotTransform(TypedDict):
        column_header: ColumnHeader
        transformation: PivotColumnTransformation

    class ExcelRangeRangeImport(TypedDict):
        type: Literal['range']
        df_name: str
        value: str

    class ExcelRangeDynamicImport(TypedDict):
        type: Literal['dynamic']
        df_name: str
        start_condition: Dict[str, Any]
        end_condition: Dict[str, Any]
        column_end_condition: Dict[str, Any]

    ExcelRangeImport = Union[ExcelRangeRangeImport, ExcelRangeDynamicImport]

    class CodeSnippet(TypedDict):
        Id: str
        Name: str
        Description: str
        Code: List[str]


    class SnowflakeCredentials(TypedDict):
        type: str
        username: str
        password: str 
        account: str

    class SnowflakeTableLocationAndWarehouseOptional(TypedDict):
        role: Optional[str]
        warehouse: Optional[str] 
        database: Optional[str]
        schema: Optional[str]
        table_or_view: Optional[str] 

    class SnowflakeTableLocationAndWarehouse(TypedDict):
        role: str
        warehouse: str
        database: str
        schema: str
        table_or_view: str

    class SnowflakeQueryParams(TypedDict):
        columns: List[str]
        limit: Optional[int]
        
    class CodeSnippetEnvVars(TypedDict):
        MITO_CONFIG_CODE_SNIPPETS_VERSION: str
        MITO_CONFIG_CODE_SNIPPETS_URL: str
        MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL: Optional[str]

    class RawParserMatch(TypedDict):
        type: Literal['{HEADER}', '{INDEX}']
        substring_range: ParserMatchSubstringRange
        unparsed: str
        parsed: Any
        row_offset: RowOffset

    class ParserMatch(TypedDict):
        type: Literal['{HEADER}', '{HEADER}{INDEX}', '{HEADER}:{HEADER}', '{HEADER}{INDEX}:{HEADER}{INDEX}']
        substring_range: ParserMatchSubstringRange
        unparsed: str
        parsed: Any
        row_offset: Union[RowOffset, Tuple[RowOffset, RowOffset]]

    class FrontendFormulaString(TypedDict):
        type: Literal['string part']
        string: str

    class FrontendFormulaHeaderIndexReference(TypedDict):
        type: Literal['{HEADER}{INDEX}']
        display_column_header: str
        row_offset: Optional[int]

    class FrontendFormulaHeaderReference(TypedDict):
        type: Literal['{HEADER}']
        display_column_header: str

    class FormulaLocationEntireColumn(TypedDict):
        type: Literal['entire_column']

    class FormulaLocationToSpecificIndexLabels(TypedDict):
        type: Literal['specific_index_labels']
        index_labels: List[Any]

    class Selection(TypedDict):
        selected_df_name: str
        selected_column_headers: List[ColumnHeader]
        selected_row_labels: List[IndexLabel]

    class DataframeReconData(TypedDict):
        created_dataframes: Dict[str, pd.DataFrame]
        deleted_dataframes: List[str]
        modified_dataframes: Dict[str, pd.DataFrame]
        last_line_expression_value: Optional[Any]
        prints: str

    class ColumnReconData(TypedDict):
        created_columns: List[ColumnHeader]
        deleted_columns: List[ColumnHeader]
        modified_columns: List[ColumnHeader]
        renamed_columns: Dict[ColumnHeader, ColumnHeader]

    class ModifiedDataframeReconData(TypedDict):
        column_recon: ColumnReconData
        num_added_or_removed_rows: int

    class AITransformFrontendResult(TypedDict):
        last_line_value: Optional[Union[str, bool, int, float, np.number]]
        created_dataframe_names: List[str]
        deleted_dataframe_names: List[str]
        modified_dataframes_recons: Dict[str, ModifiedDataframeReconData]
        prints: str

    class CodeOptions(TypedDict):
        as_function: bool
        call_function: bool
        function_name: str
        function_params: Dict[ParamName, ParamValue]

    UserDefinedImporterParamType = Literal['any', 'str', 'int', 'float', 'bool']

else:
    Filter = Any #type: ignore
    FilterGroup = Any #type: ignore
    OperatorType = Any #type:ignore
    PrimitiveTypeName = None # type: ignore
    ColumnIDWithFilter = Any # type:ignore
    ColumnIDWithFilterGroup = Any # type:ignore
    ColumnHeaderWithFilter = Any # type:ignore
    ColumnIDWithPivotTransform = Any # type:ignore
    ColumnHeaderWithPivotTransform = Any # type:ignore
    ExcelRangeImport = Any # type:ignore
    CodeSnippet = Any # type:ignore
    CodeSnippetEnvVars = Any # type:ignore
    RawParserMatch = Any # type:ignore
    ParserMatch = Any # type:ignore
    SnowflakeCredentials = Any # type:ignore
    SnowflakeTableLocationAndWarehouse = Any # type:ignore
    SnowflakeTableLocationAndWarehouseOptional = Any #type:ignore
    SnowflakeQueryParams = Any # type:ignore
    CodeSnippetEnvVars = Any # type:ignore
    FrontendFormulaString = Any # type:ignore
    FrontendFormulaHeaderIndexReference = Any # type:ignore
    FrontendFormulaHeaderReference = Any # type:ignore
    FormulaLocationEntireColumn = Any # type:ignore
    FormulaLocationToSpecificIndexLabels = Any # type:ignore
    Selection = Any # type:ignore
    DataframeReconData = Any # type: ignore
    ColumnReconData = Any # type: ignore
    ModifiedDataframeReconData = Any # type: ignore
    AITransformFrontendResult = Any # type: ignore
    CodeOptions = Any # type: ignore
    UserDefinedImporterParamType = Any # type: ignore


FrontendFormulaPart = Union[FrontendFormulaString, FrontendFormulaHeaderIndexReference, FrontendFormulaHeaderReference]
FrontendFormula = List[FrontendFormulaPart]

FormulaAppliedToType = Union[FormulaLocationEntireColumn, FormulaLocationToSpecificIndexLabels]


if sys.version_info[:3] > (3, 8, 0):
    from typing import TypedDict

    class FrontendFormulaAndLocation(TypedDict):
        frontend_formula: FrontendFormula
        location: FormulaAppliedToType
        index: List[Any]

else:
    FrontendFormulaAndLocation = Any # type:ignore






import string
from typing import List, Optional, Tuple, Union

import pandas as pd

from mitosheet.transpiler.transpile_utils import (
    CLOSE_BRACKET, NEWLINE, OPEN_BRACKET,
    column_header_list_to_transpiled_code, column_header_to_transpiled_code)
from mitosheet.types import Selection

MAX_TOKENS = 4096
CHARS_PER_TOKEN = 4 # https://platform.openai.com/docs/introduction/tokens
MAX_CHARS = MAX_TOKENS * CHARS_PER_TOKEN

MAX_TOKENS_FOR_INPUT_DATA = 3800
MAX_CHARS_FOR_INPUT_DATA = MAX_TOKENS_FOR_INPUT_DATA * CHARS_PER_TOKEN

PROMPT_VERSION = 'df-creation-prompt-1'

def get_dataframe_creation_code(df: pd.DataFrame, max_characters: Union[int, float]) -> str:
    """
    Given a dataframe like:

        date (year)      value sum
    0         2000             1.0
    1         2001             NaN

    Will return a string that looks something like
    ```
    pd.DataFrame({'date (year)': [2000, 2001], 'value sum: [1.0, np.NaN]})
    ```

    This notably will only include max_characters in the string it returns.
    It will chop characters first by not including the data present in the
    dataframe and just including the columns. 

    Then, it will chop characters by just including the column names, and 
    note the values in the columns. Then, it will chop characters by eliding
    some of the columns (it just takes a greedy approach to taking columns).
    """

    transpiled_column_headers_to_transpiled_values = {
        column_header_to_transpiled_code(column_header): column_header_list_to_transpiled_code(df[column_header].to_list())
        for column_header in df.columns
    }

    column_headers_to_values_string = ", ".join([f"{ch}: {vs}" for ch, vs in transpiled_column_headers_to_transpiled_values.items()])
    df_string_with_values = f'pd.DataFrame({OPEN_BRACKET}{column_headers_to_values_string}{CLOSE_BRACKET})'

    if len(df_string_with_values) < max_characters:
        return df_string_with_values
    
    column_headers_string = ", ".join([ch for ch in transpiled_column_headers_to_transpiled_values])
    df_string_with_only_headers = f'pd.DataFrame(columns=[{column_headers_string}])'

    if len(df_string_with_only_headers) < max_characters:
        return df_string_with_only_headers

    # Then, we find the maximum number of column headers we can include. Note that we use a binary search algorithm
    # to do this - as the loop that we original started with had performance problems on dataframes with a large
    # number of column headers that we noticed while testing
    reduced_headers = ''
    columns = df.columns.to_list()
    low = 0
    num_not_included = 1
    high = len(columns) - 1
    len_container = len(f'pd.DataFrame(columns=[ ... and {high} more])') # Max length of container

    while low < high:
        mid = (low + high) // 2
        reduced_headers = column_header_list_to_transpiled_code(columns[:mid])
        num_not_included = len(columns) - mid + 1

        if (len(reduced_headers) + len_container) < max_characters:
            low = mid + 1
        elif (len(reduced_headers) + len_container) > max_characters:
            high = mid - 1

    return f'pd.DataFrame({OPEN_BRACKET}columns=[{reduced_headers} ... and {num_not_included} more]{CLOSE_BRACKET})'


def get_input_data_string(df_names: List[str], dfs: List[pd.DataFrame], current_selection: Optional[Selection]) -> str:
    # First, we get the input data code for creating the dataframes
    dfs_creation_code = {}
    for df_name, df in zip(df_names, dfs):
        max_characters = MAX_CHARS_FOR_INPUT_DATA / len(df_names)
        dfs_creation_code[df_name] = get_dataframe_creation_code(df.head(5), max_characters)

    # Then, build the input data string, making sure to put the selected dataframe first - if these is one
    if current_selection is not None:
        input_data_string = f"""Input dataframes:
```
{current_selection['selected_df_name']} = {dfs_creation_code[current_selection['selected_df_name']]}
{NEWLINE.join([f'{df_name} = {df_creation_code}' for df_name, df_creation_code in dfs_creation_code.items() if df_name != current_selection['selected_df_name']])}
```
The dataframe {current_selection['selected_df_name']} is currently selected.
"""
    else:
        input_data_string = f"""Input dataframes:
```
{NEWLINE.join([f'{df_name} = {df_creation_code}' for df_name, df_creation_code in dfs_creation_code.items()])}
```"""

    return input_data_string

def get_transformation_and_code_string(transformation: str, code: Optional[str]=None) -> str:
    if code is None:
        return f"""Transformation: {transformation}
Code:
```
"""
    else:
        return f"""Transformation: {transformation}
Code:
```
{code}
```"""


def get_example_string(df_names: List[str], current_selection: Optional[Selection]) -> str:

    # TODO: If there is a previous edit in the mitosheet, we could use this as the example, as it's about as in-context as you can get

    # Otherwise, we make up a fake edit
    if current_selection is not None:
        return get_transformation_and_code_string(
            f"replace 1 with 2",
            f"{current_selection['selected_df_name']}.replace(1, 2, inplace=True)"
        )
    else:
        df_name = df_names[0] if len(df_names) > 0 else 'df'
        return get_transformation_and_code_string(
            f"replace 1 with 2 in {df_name}",
            f"{df_name}.replace(1, 2, inplace=True)"
        )


def get_previous_failed_completions_string(previous_failed_completions: List[Tuple[str, str]], user_input: str) -> str:
    if len(previous_failed_completions) == 0:
        return ""
    
    final_str = ''
    for index, (code, error) in enumerate(previous_failed_completions):

        if index == 0:
            final_str += f"""{code}
```
Error:
```
{error}
```

"""
        else:
            final_str += f"""Fixing previous errors, {get_transformation_and_code_string(user_input, code)}
Error:
```
{error}
```

"""

                
    final_str += f"Fixing previous errors, {get_transformation_and_code_string(user_input)}"

    return final_str


def get_prompt(
        df_names: List[str], 
        dfs: List[pd.DataFrame], 
        current_selection: Optional[Selection], 
        user_input: str,
        previous_failed_completions: List[Tuple[str, str]]
    ) -> str:

    previous_failed_completions_string = get_previous_failed_completions_string(previous_failed_completions, user_input)

    if len(dfs) > 0:
        input_data_string = get_input_data_string(df_names, dfs, current_selection)
        example_string = get_example_string(df_names, current_selection)

        return f"""You are a pandas developer who is working with the following dataframes.

{input_data_string}

{example_string}

{get_transformation_and_code_string(user_input)}{previous_failed_completions_string}
"""
    else:

        return f"""
You are a pandas developer who has just started a new script. You are looking to 
make the following transformation.

{get_transformation_and_code_string(user_input)}{previous_failed_completions_string}
"""



import streamlit as st




st.set_page_config(layout="wide")

from mitosheet.streamlit.v1 import spreadsheet
new_dfs, code = spreadsheet('https://raw.githubusercontent.com/plotly/datasets/master/tesla-stock-price.csv')
st.code(code)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import List, Tuple
from mitosheet.ai.ai_utils import get_import_lines_to_add
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State

class AITransformationCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, user_input: str, final_code: str):
        super().__init__(prev_state, post_state)
        self.user_input = user_input
        self.final_code = final_code

    def get_display_name(self) -> str:
        return 'AI Transformation'
    
    def get_description_comment(self) -> str:
        return self.user_input

    def get_code(self) -> Tuple[List[str], List[str]]:
        import_lines = get_import_lines_to_add(self.final_code)
        return [
            self.final_code.strip()
        ], import_lines

    def get_edited_sheet_indexes(self) -> List[int]:
        # TODO: we can save this perhaps?
        return [] # TODO: return this here!
    



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Dict, List, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import TAB, column_header_to_transpiled_code

from mitosheet.transpiler.transpile_utils import param_dict_to_code

from mitosheet.utils import (
    get_conditional_formats_objects_to_export_to_excel, 
    get_number_formats_objects_to_export_to_excel
)

# This is a helper function that generates the code for formatting the excel sheet
def get_format_code(state: State, sheet_index_to_export_location: Dict[int, str]) -> list:
    code = []
    formats = state.df_formats
    for sheet_index, export_location in sheet_index_to_export_location.items():
        format = formats[sheet_index]
        # We need to convert the column IDs to column letters
        # for conditional formats to export to excel
        conditional_formats = get_conditional_formats_objects_to_export_to_excel(
            format.get('conditional_formats'),
            column_id_map=state.column_ids,
            sheet_index=sheet_index
        )
        number_formats = get_number_formats_objects_to_export_to_excel(format.get('columns'))
        params = {
            'header_background_color': format.get('headers', {}).get('backgroundColor'),
            'header_font_color': format.get('headers', {}).get('color'),
            'even_background_color': format.get('rows', {}).get('even', {}).get('backgroundColor'),
            'even_font_color': format.get('rows', {}).get('even', {}).get('color'),
            'odd_background_color': format.get('rows', {}).get('odd', {}).get('backgroundColor'),
            'odd_font_color': format.get('rows', {}).get('odd', {}).get('color'),
            'conditional_formats': conditional_formats,
            'number_formats': number_formats,
        }
        param_dict = {
            key: value for key, value in params.items()
            if value is not None
        }
        if param_dict == {}:
            continue

        params_code = param_dict_to_code(param_dict, tab_level=1)
        code.append(f'{TAB}add_formatting_to_excel_sheet(writer, "{export_location}", {state.df_names[sheet_index]}, {params_code})')
    return code


class ExportToFileCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, export_type: str, file_name: str, sheet_index_to_export_location: Dict[int, str], export_formatting: bool=False):
        super().__init__(prev_state, post_state)
        self.export_type = export_type
        self.file_name = file_name
        self.sheet_index_to_export_location = sheet_index_to_export_location
        self.export_formatting = export_formatting

    def get_display_name(self) -> str:
        return 'Export To File'

    def get_description_comment(self) -> str:
        return f"Exports {len(self.sheet_index_to_export_location)} to file {self.file_name}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        if self.export_type == 'csv':
            return [
                f"{self.post_state.df_names[sheet_index]}.to_csv(r{column_header_to_transpiled_code(export_location)}, index=False)"
                for sheet_index, export_location in self.sheet_index_to_export_location.items()
            ], []
        elif self.export_type == 'excel':
            format_code = get_format_code(self.post_state, self.sheet_index_to_export_location) if self.export_formatting else []
            return [f"with pd.ExcelWriter(r{column_header_to_transpiled_code(self.file_name)}, engine=\"openpyxl\") as writer:"] + [
                f'{TAB}{self.post_state.df_names[sheet_index]}.to_excel(writer, sheet_name="{export_location}", index={False})'
                for sheet_index, export_location in self.sheet_index_to_export_location.items()
            ] + format_code, ['import pandas as pd']
        else:
            raise ValueError(f'Not a valid file type: {self.export_type}')
        
    def get_parameterizable_params(self) -> List[Tuple[str, str, str]]:
        if self.export_type == 'csv':
            return [
                (f"r{column_header_to_transpiled_code(export_location)}", 'file_name', 'CSV export file path') for export_location in self.sheet_index_to_export_location.values()
            ]
        elif self.export_type == 'excel':
            return [(f"r{column_header_to_transpiled_code(self.file_name)}", 'file_name', 'Excel export file path')]
        else:
            raise ValueError(f'Not a valid file type: {self.export_type}')
